from bluesky_live.run_builder import build_simple_run
import pytest

from ..plot_builders import Lines
from ..plot_specs import AxesSpec, FigureSpec
from ...headless.figures import HeadlessFigure


# Make some runs to use.
runs = [
    build_simple_run(
        {"motor": [1, 2], "det": [10, 20], "det2": [15, 25]},
        metadata={"scan_id": 1 + i},
    )
    for i in range(10)
]
MAX_RUNS = 3


def test_pinned():
    "Test Lines with 'pinned' and un-pinned runs."
    ys = ["det", "det2"]
    num_ys = len(ys)
    model = Lines("motor", ys, max_runs=MAX_RUNS)
    view = HeadlessFigure(model.figure)

    # Add MAX_RUNS and then some more and check that they do get bumped off.
    for run in runs[:5]:
        model.add_run(run)
        assert len(model.runs) <= MAX_RUNS
    assert runs[2:5] == list(model.runs)

    # Add a pinned run.
    pinned_run = runs[5]
    model.add_run(pinned_run, pinned=True)
    assert frozenset([pinned_run.metadata["start"]["uid"]]) == model.pinned
    for run in runs[6:]:
        model.add_run(run)
        assert len(model.runs) == 1 + MAX_RUNS
        assert len(model.figure.axes[0].lines) == num_ys * (1 + MAX_RUNS)
    # Check that it hasn't been bumped off.
    assert pinned_run in model.runs

    # Remove the pinned run.
    model.discard_run(pinned_run)
    assert len(model.runs) == MAX_RUNS
    assert len(model.figure.axes[0].lines) == num_ys * MAX_RUNS
    assert pinned_run not in model.runs

    view.close()


def test_properties():
    "Touch various accessors"
    model = Lines("c * motor", ["det"], namespace={"c": 3}, max_runs=MAX_RUNS)
    view = HeadlessFigure(model.figure)
    model.add_run(runs[0])
    assert model.runs[0] is runs[0]
    assert model.max_runs == MAX_RUNS
    assert model.x == "c * motor"
    assert model.ys == ("det",)
    assert dict(model.namespace) == {"c": 3}
    assert model.needs_streams == ("primary",)
    assert model.pinned == frozenset()

    view.close()


def test_decrease_max_runs():
    "Decreasing max_runs should remove the runs and their associated lines."
    INITIAL_MAX_RUNS = 5
    model = Lines("motor", ["det"], namespace={"c": 3}, max_runs=INITIAL_MAX_RUNS)
    view = HeadlessFigure(model.figure)
    for run in runs[:5]:
        model.add_run(run)
    assert len(model.runs) == INITIAL_MAX_RUNS
    assert len(model.figure.axes[0].lines) == INITIAL_MAX_RUNS
    # Decrease max_runs.
    model.max_runs = MAX_RUNS
    assert len(model.runs) == MAX_RUNS
    assert len(model.figure.axes[0].lines) == MAX_RUNS

    view.close()


@pytest.mark.parametrize("expr", ["det / det2", "-log(det)", "np.sqrt(det)"])
def test_expressions(expr):
    "Test Lines with 'pinned' and un-pinned runs."
    ys = [expr]
    model = Lines("motor", ys, max_runs=MAX_RUNS)
    view = HeadlessFigure(model.figure)
    model.add_run(runs[0])
    assert len(model.figure.axes[0].lines) == 1

    view.close()


@pytest.mark.parametrize(
    "func",
    [
        lambda det, det2: det / det2,
        lambda det, log: -log(det),
        lambda det, np: np.sqrt(det),
    ],
)
def test_functions(func):
    "Test Lines with 'pinned' and un-pinned runs."
    ys = [func]
    model = Lines("motor", ys, max_runs=MAX_RUNS)
    view = HeadlessFigure(model.figure)
    model.add_run(runs[0])
    assert len(model.figure.axes[0].lines) == 1

    view.close()


def test_figure_set_after_instantiation():
    axes = AxesSpec()
    model = Lines("motor", [], axes=axes)
    assert model.figure is None
    figure = FigureSpec((axes,), title="")
    assert model.figure is figure
    view = HeadlessFigure(model.figure)
    view.close()



from ..plot_builders import Images
from ._base import AutoPlotter


class AutoImages(AutoPlotter):
    """
    Construct figures with line plots automatically.

    The decision of which lines to plot is based on metadata and data shape.

    Examples
    --------

    View with Jupyter.

    >>> model = AutoImages()
    >>> from bluesky_widgets.jupyter.figures import JupyterFigures
    >>> view = JupyterFigures(model.figures)
    >>> model.add_run(run)
    """

    def __init__(self, *, max_runs=None):
        super().__init__()
        # Map (stream_name, field) to instance of Images
        self._field_to_builder = {}
        self._max_runs = max_runs

    @property
    def max_runs(self):
        return self._max_runs

    @max_runs.setter
    def max_runs(self, value):
        if value is not None:
            for builders in self._lines_instances.values():
                for builder in builders:
                    builder.max_runs = value
        self._max_runs = value

    def handle_new_stream(self, run, stream_name):
        """
        This is used internally and should not be called directly by user code.

        Given a run and stream name, add or update figures with images.

        Parameters
        ----------
        run : BlueskyRun
        stream_name : String
        """
        ds = run[stream_name].to_dask()
        for field in ds:
            if 2 <= ds[field].ndim < 5:
                try:
                    images = self._field_to_builder[(stream_name, field)]
                except KeyError:
                    images = Images(field=field, needs_streams=(stream_name,))
                    self._field_to_builder[(stream_name, field)] = images
                images.add_run(run)
                self.plot_builders.append(images)
                self.figures.append(images.figure)



from bluesky_live.run_builder import RunBuilder
import pytest
import numpy

from ..plot_builders import RasteredImages
from ..plot_specs import AxesSpec, FigureSpec


@pytest.fixture
def non_snaking_run():
    # Test data
    md = {"motors": ["y", "x"], "shape": [2, 2], "snaking": (False, False)}
    with RunBuilder(md) as builder:
        builder.add_stream(
            "primary", data={"ccd": [1, 2, 3, 4], "x": [0, 1, 0, 1], "y": [0, 0, 1, 1]}
        )
    run = builder.get_run()
    return run


@pytest.fixture
def snaking_run():
    # Test data
    md = {"motors": ["y", "x"], "shape": [2, 2], "snaking": (False, True)}
    with RunBuilder(md) as builder:
        builder.add_stream(
            "primary", data={"ccd": [1, 2, 3, 4], "x": [0, 1, 1, 0], "y": [0, 0, 1, 1]}
        )
    run = builder.get_run()
    return run


def test_rastered_image(non_snaking_run):
    "Test RasteredImages with a 2D array."
    run = non_snaking_run
    model = RasteredImages("ccd", shape=(2, 2))
    assert not model.figure.axes[0].images
    model.add_run(run)
    assert model.figure.axes[0].images


def test_x_y_positive_change_x_y_limits(non_snaking_run):
    "Test x_positive and y_positive change x_limits and y_limits"
    run = non_snaking_run
    model = RasteredImages("ccd", shape=(2, 2), x_positive="left", y_positive="down")
    model.add_run(run)
    expected_x_lims = expected_y_lims = (1.5, -0.5)
    assert model.axes.x_limits == expected_x_lims
    assert model.axes.y_limits == expected_y_lims
    model.x_positive = "right"
    model.y_positive = "up"
    expected_x_lims = expected_y_lims = (-0.5, 1.5)
    assert model.axes.x_limits == expected_x_lims
    assert model.axes.y_limits == expected_y_lims


def test_x_y_limits_change_x_y_positive(non_snaking_run):
    "Test x_limits and y_limits change x_positive and y_positive"
    run = non_snaking_run
    axes = AxesSpec(x_limits=(1.5, -0.5), y_limits=(1.5, -0.5))
    model = RasteredImages("ccd", shape=(2, 2), axes=axes)
    model.add_run(run)
    assert model.x_positive == "left"
    assert model.y_positive == "down"
    model.axes.x_limits = model.axes.y_limits = (-0.5, 1.5)
    assert model.x_positive == "right"
    assert model.y_positive == "up"


def test_non_snaking_image_data(non_snaking_run):
    run = non_snaking_run
    model = RasteredImages("ccd", shape=(2, 2))
    model.add_run(run)
    actual_data = model.figure.axes[0].images[0].func(run)
    expected_data = [[1, 2], [3, 4]]
    assert numpy.array_equal(actual_data, expected_data)


def test_snaking_image_data(snaking_run):
    run = snaking_run
    model = RasteredImages("ccd", shape=(2, 2))
    model.add_run(run)
    actual_data = model.figure.axes[0].images[0].func(run)
    expected_data = [[1, 2], [4, 3]]
    assert numpy.array_equal(actual_data, expected_data)


def test_non_snaking_image_data_positions():
    md = {"motors": ["y", "x"], "shape": [2, 2], "snaking": (False, False)}
    model = RasteredImages("ccd", shape=(2, 2))
    with RunBuilder(md) as builder:
        ccd = iter([1, 2, 3, 4])
        x = iter([0, 1, 0, 1])
        y = iter([0, 0, 1, 1])
        run = builder.get_run()
        model.add_run(run)
        # First data point
        builder.add_stream(
            "primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]}
        )
        actual_data = model.figure.axes[0].images[0].func(run)
        expected_data = [[1, numpy.nan], [numpy.nan, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Second point
        builder.add_data(
            "primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]}
        )
        actual_data = model.figure.axes[0].images[0].func(run)
        expected_data = [[1, 2], [numpy.nan, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Third point
        builder.add_data(
            "primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]}
        )
        actual_data = model.figure.axes[0].images[0].func(run)
        expected_data = [[1, 2], [3, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Fourth point
        builder.add_data(
            "primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]}
        )
        actual_data = model.figure.axes[0].images[0].func(run)
        expected_data = [[1, 2], [3, 4]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)


def test_snaking_image_data_positions():
    md = {"motors": ["y", "x"], "shape": [2, 2], "snaking": (False, True)}
    model = RasteredImages("ccd", shape=(2, 2))
    with RunBuilder(md) as builder:
        ccd = iter([1, 2, 3, 4])
        x = iter([0, 1, 1, 0])
        y = iter([0, 0, 1, 1])
        run = builder.get_run()
        model.add_run(run)
        # First data point
        builder.add_stream(
            "primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]}
        )
        actual_data = model.figure.axes[0].images[0].func(run)
        expected_data = [[1, numpy.nan], [numpy.nan, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Second point
        builder.add_data(
            "primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]}
        )
        actual_data = model.figure.axes[0].images[0].func(run)
        expected_data = [[1, 2], [numpy.nan, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Third point
        builder.add_data(
            "primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]}
        )
        actual_data = model.figure.axes[0].images[0].func(run)
        expected_data = [[1, 2], [numpy.nan, 3]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Fourth point
        builder.add_data(
            "primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]}
        )
        actual_data = model.figure.axes[0].images[0].func(run)
        expected_data = [[1, 2], [4, 3]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)


def test_figure_set_after_instantiation():
    axes = AxesSpec()
    model = RasteredImages("ccd", shape=(2, 2), axes=axes)
    assert model.figure is None
    figure = FigureSpec((axes,), title="")
    assert model.figure is figure



"""
Run like:

python -m bluesky_widgets.examples.kafka_figures

For each Run, it will generate thumbnails and save them to a temporary
directory. The filepaths will be printed to the stdout, one per line.
"""
from functools import partial
import os
import tempfile

import msgpack
import msgpack_numpy as mpn
from bluesky_kafka import RemoteDispatcher

from bluesky_widgets.utils.streaming import stream_documents_into_runs
from bluesky_widgets.models.plot_builders import AutoLines
from bluesky_widgets.headless.figures import HeadlessFigures
from bluesky_widgets.models.utils import run_is_live_and_not_completed


def export_thumbnails_when_complete(run):
    "Given a BlueskyRun, export thumbnail(s) to a directory when it completes."
    model = AutoLines(max_runs=3)
    model.add_run(run)
    view = HeadlessFigures(model.figures)

    uid = run.metadata["start"]["uid"]
    directory = os.path.join(tempfile.gettempdir(), "bluesky_widgets_example", uid)
    os.makedirs(directory, exist_ok=True)

    # If the Run is already done by the time we got it, export now.
    # Otherwise, schedule it to export whenever it finishes.
    def export(*args):
        filenames = view.export_all(directory)
        print("\n".join(f'"{filename}"' for filename in filenames))
        view.close()

    if run_is_live_and_not_completed(run):
        run.events.new_data.connect(export)
    else:
        export()


if __name__ == "__main__":
    bootstrap_servers = "127.0.0.1:9092"
    kafka_deserializer = partial(msgpack.loads, object_hook=mpn.decode)
    topics = ["widgets_test.bluesky.documents"]
    consumer_config = {"auto.commit.interval.ms": 100, "auto.offset.reset": "latest"}

    dispatcher = RemoteDispatcher(
        topics=topics,
        bootstrap_servers=bootstrap_servers,
        group_id="widgets_test",
        consumer_config=consumer_config,
    )

    dispatcher.subscribe(stream_documents_into_runs(export_thumbnails_when_complete))
    dispatcher.start()



import pytest

from qtpy.QtWidgets import QApplication


@pytest.fixture
def qtbot(qtbot):
    """A modified qtbot fixture that makes sure no widgets have been leaked."""
    # Adapted from napari
    initial = QApplication.topLevelWidgets()
    yield qtbot
    QApplication.processEvents()
    leaks = set(QApplication.topLevelWidgets()).difference(initial)
    if leaks:
        # Ignore this until we resolve
        # AssertionError: Widgets leaked!:
        # {matplotlib.backends.backend_qt5.MainWindow object at 0x7fdce80b7550>}
        # raise AssertionError(f"Widgets leaked!: {leaks}")
        pass



from bluesky_live.run_builder import RunBuilder
import pytest

from ...models.plot_specs import (
    FigureSpec,
    FigureSpecList,
    AxesSpec,
    LineSpec,
    ImageSpec,
)
from ..figures import QtFigure, QtFigures


# Generate example data.
with RunBuilder() as builder:
    builder.add_stream("primary", data={"a": [1, 2, 3], "b": [1, 4, 9]})
run = builder.get_run()


def transform(run):
    ds = run.primary.read()
    return ds["a"], ds["b"]


def func(run):
    line = LineSpec(transform, run, "label")
    axes = AxesSpec(lines=[line], x_label="a", y_label="b", title="axes title")
    figure = FigureSpec((axes,), title="figure title")
    return figure


def test_figure(qtbot):
    "Basic test: create a QtFigure."
    figure = func(run)
    QtFigure(figure)


def test_figures(qtbot):
    "Basic test: create QtFigures."
    figure = func(run)
    another_figure = func(run)
    figures = FigureSpecList([figure, another_figure])
    QtFigures(figures)


def test_figure_title_syncing(qtbot):
    model = func(run)
    view = QtFigure(model)
    initial = model.title
    assert view.figure._suptitle.get_text() == initial
    expected = "figure title changed"
    model.title = expected
    assert model.title == expected
    # Is there no public matplotlib API for this?
    # https://stackoverflow.com/a/48917679
    assert view.figure._suptitle.get_text() == expected


def test_short_title_syncing(qtbot):
    model = func(run)
    figures = FigureSpecList([model])
    view = QtFigures(figures)
    actual_title = view.figures[model.uuid].figure._suptitle.get_text()
    assert view.tabText(0) == actual_title
    expected_short_title = "new short title"
    model.short_title = expected_short_title
    assert model.short_title == expected_short_title
    actual_title = view.figures[model.uuid].figure._suptitle.get_text()
    assert view.tabText(0) == expected_short_title
    assert actual_title == model.title
    expected_title = "new title"
    model.title = expected_title
    assert view.tabText(0) == model.short_title
    model.short_title = None
    assert view.tabText(0) == expected_title


def test_non_null_short_title_syncing(qtbot):
    model = func(run)
    model.short_title = "short title"
    figures = FigureSpecList([model])
    view = QtFigures(figures)
    actual_title = view.figures[model.uuid].figure._suptitle.get_text()
    assert view.tabText(0) == model.short_title
    assert actual_title == model.title


@pytest.mark.parametrize(
    ("model_property", "mpl_method"),
    [("title", "get_title"), ("x_label", "get_xlabel"), ("y_label", "get_ylabel")],
)
def test_axes_syncing(qtbot, model_property, mpl_method):
    model = func(run)
    view = QtFigure(model)
    initial = getattr(model.axes[0], model_property)
    assert getattr(view.figure.axes[0], mpl_method)() == initial
    expected = "axes title changed"
    setattr(model.axes[0], model_property, expected)
    assert getattr(model.axes[0], model_property) == expected
    assert getattr(view.figure.axes[0], mpl_method)() == expected


def test_axes_set_figure():
    "Adding axes to a figure sets their figure."
    axes = AxesSpec()
    assert axes.figure is None
    figure = FigureSpec((axes,), title="figure title")
    assert axes.figure is figure
    # Once axes belong to a figure, they cannot belong to another figure.
    with pytest.raises(RuntimeError):
        FigureSpec((axes,), title="figure title")

    with pytest.raises(AttributeError):
        figure.axes = (axes,)  # not settable


artist_set_axes_params = pytest.mark.parametrize(
    ("model_property", "artist_factory"),
    [
        ("lines", lambda: LineSpec(transform, run, "label")),
        ("images", lambda: ImageSpec(transform, run, "label")),
    ],
    ids=["lines", "images"],
)


@artist_set_axes_params
def test_artist_set_axes_at_init(model_property, artist_factory):
    "Adding an artist to axes at init time sets its axes."
    artist = artist_factory()
    axes = AxesSpec(**{model_property: [artist]})
    assert artist in getattr(axes, model_property)
    assert artist in axes.by_uuid.values()
    assert artist.axes is axes

    # Once line belong to a axes, it cannot belong to another axes.
    with pytest.raises(RuntimeError):
        AxesSpec(**{model_property: [artist]})


@artist_set_axes_params
def test_artist_set_axes_after_init(model_property, artist_factory):
    "Adding an artist to axes after init time sets its axes."
    artist = artist_factory()
    axes = AxesSpec()
    getattr(axes, model_property).append(artist)
    assert artist in getattr(axes, model_property)
    assert artist in axes.by_uuid.values()
    assert artist.axes is axes

    # Once line belong to a axes, it cannot belong to another axes.
    with pytest.raises(RuntimeError):
        AxesSpec(**{model_property: [artist]})



"""
Custom Qt widgets that serve as native objects that the public-facing elements
wrap.
"""
import time

from qtpy.QtWidgets import (  # noqa: E402
    QApplication,
    QMainWindow,
    QWidget,
    QHBoxLayout,
    QLabel,
    QStatusBar,
    QFileDialog,
)
from qtpy.QtCore import Qt  # noqa: E402

from ._event_loop import get_our_app_name
from .threading import wait_for_workers_to_quit


class Window:
    """Application window that contains the menu bar and viewer.

    Parameters
    ----------
    qt_widget : QtViewer
        Contained viewer widget.

    Attributes
    ----------
    file_menu : qtpy.QtWidgets.QMenu
        File menu.
    help_menu : qtpy.QtWidgets.QMenu
        Help menu.
    main_menu : qtpy.QtWidgets.QMainWindow.menuBar
        Main menubar.
    qt_widget : QtViewer
        Contained viewer widget.
    view_menu : qtpy.QtWidgets.QMenu
        View menu.
    window_menu : qtpy.QtWidgets.QMenu
        Window menu.
    """

    def __init__(self, qt_widget, *, show):

        self.qt_widget = qt_widget

        self._qt_window = QMainWindow()
        self._qt_window.setAttribute(Qt.WA_DeleteOnClose)
        self._qt_window.setUnifiedTitleAndToolBarOnMac(True)
        self._qt_center = QWidget(self._qt_window)

        self._qt_window.setCentralWidget(self._qt_center)
        if hasattr(self.qt_widget.model, "title"):
            self._qt_window.setWindowTitle(self.qt_widget.model.title)
        self._qt_center.setLayout(QHBoxLayout())
        self._status_bar = QStatusBar()
        self._qt_window.setStatusBar(self._status_bar)

        self._status_bar.showMessage("Ready")
        self._help = QLabel("")
        self._status_bar.addPermanentWidget(self._help)

        self._qt_center.layout().addWidget(self.qt_widget)
        self._qt_center.layout().setContentsMargins(4, 0, 4, 0)

        # self._add_viewer_dock_widget(self.qt_widget.dockConsole)
        # self._add_viewer_dock_widget(self.qt_widget.dockLayerControls)
        # self._add_viewer_dock_widget(self.qt_widget.dockLayerList)

        # self.qt_widget.viewer.events.status.connect(self._status_changed)
        # self.qt_widget.viewer.events.help.connect(self._help_changed)
        # self.qt_widget.viewer.events.title.connect(self._title_changed)
        # self.qt_widget.viewer.events.palette.connect(self._update_palette)

        if show:
            self.show()

    def resize(self, width, height):
        """Resize the window.

        Parameters
        ----------
        width : int
            Width in logical pixels.
        height : int
            Height in logical pixels.
        """
        self._qt_window.resize(width, height)

    def show(self):
        """Resize, show, and bring forward the window."""
        self._qt_window.resize(self._qt_window.layout().sizeHint())
        self._qt_window.show()

        # We want to call Window._qt_window.raise_() in every case *except*
        # when instantiating a viewer within a gui_qt() context for the
        # _first_ time within the Qt app's lifecycle.
        #
        # `app_name` will be ours iff the application was instantiated in
        # gui_qt(). isActiveWindow() will be True if it is the second time a
        # _qt_window has been created. See #732
        app_name = QApplication.instance().applicationName()
        if app_name != get_our_app_name() or self._qt_window.isActiveWindow():
            self._qt_window.raise_()  # for macOS
            self._qt_window.activateWindow()  # for Windows

    def _status_changed(self, event):
        """Update status bar.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self._status_bar.showMessage(event.text)

    def _title_changed(self, event):
        """Update window title.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self._qt_window.setWindowTitle(event.text)

    def _help_changed(self, event):
        """Update help message on status bar.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self._help.setText(event.text)

    def _screenshot_dialog(self):
        """Save screenshot of current display with viewer, default .png"""
        filename, _ = QFileDialog.getSaveFileName(
            parent=self.qt_widget,
            caption="Save screenshot with viewer",
            directory=self.qt_widget._last_visited_dir,  # home dir by default
            filter="Image files (*.png *.bmp *.gif *.tif *.tiff)",  # first one used by default
            # jpg and jpeg not included as they don't support an alpha channel
        )
        if (filename != "") and (filename is not None):
            # double check that an appropriate extension has been added as the
            # filter option does not always add an extension on linux and windows
            # see https://bugreports.qt.io/browse/QTBUG-27186
            image_extensions = (".bmp", ".gif", ".png", ".tif", ".tiff")
            if not filename.endswith(image_extensions):
                filename = filename + ".png"
            self.screenshot(path=filename)

    def screenshot(self, path=None):
        """Take currently displayed viewer and convert to an image array.

        Parameters
        ----------
        path : str
            Filename for saving screenshot image.

        Returns
        -------
        image : array
            Numpy array of type ubyte and shape (h, w, 4). Index [0, 0] is the
            upper-left corner of the rendered region.
        """
        img = self._qt_window.grab().toImage()
        if path is not None:
            from skimage.io import imsave
            from .utils import QImg2array  # noqa: E402

            imsave(path, QImg2array(img))
        return QImg2array(img)

    def close(self):
        """Close the viewer window and cleanup sub-widgets."""
        # on some versions of Darwin, exiting while fullscreen seems to tickle
        # some bug deep in NSWindow.  This forces the fullscreen keybinding
        # test to complete its draw cycle, then pop back out of fullscreen.
        if self._qt_window.isFullScreen():
            self._qt_window.showNormal()
            for i in range(8):
                time.sleep(0.1)
                QApplication.processEvents()
        self.qt_widget.close()
        self._qt_window.close()
        wait_for_workers_to_quit()
        del self._qt_window



# Based on databroker/examples/generate_mongo_data.py
from suitcase.mongo_normalized import Serializer
from bluesky import RunEngine
from bluesky.plans import count, scan
from ophyd.sim import det, motor, SynSignal
import numpy as np
import uuid

from databroker._drivers.mongo_normalized import BlueskyMongoCatalog


random_img = SynSignal(func=lambda: np.random.random((5, 10, 10)), name="random_img")


def get_catalog():

    RE = RunEngine()

    mds = f"mongodb://localhost:27017/databroker-test-{uuid.uuid4()}"
    fs = f"mongodb://localhost:27017/databroker-test-{uuid.uuid4()}"
    serializer = Serializer(mds, fs)
    for i in range(1, 5):
        RE(scan([det], motor, -1, 1, 5 * i), serializer)
    RE(count([random_img], 3), serializer)

    catalog = BlueskyMongoCatalog(mds, fs)
    return catalog



import collections.abc

from ipywidgets import widgets

from ..models.plot_specs import FigureSpec, FigureSpecList
from .._matplotlib_axes import MatplotlibAxes
from ..utils.dict_view import DictView


def _initialize_mpl():
    "Set backend to ipympl and import pyplot."
    import matplotlib

    matplotlib.use(
        "module://ipympl.backend_nbagg"
    )  # must set before importing matplotlib.pyplot
    # must import matplotlib.pyplot here because bluesky.utils.during_task
    # expects it to be imported
    import matplotlib.pyplot as plt  # noqa


class JupyterFigures(widgets.Tab):
    """
    A Jupyter (ipywidgets) view for a FigureSpecList model.
    """

    def __init__(self, model: FigureSpecList, *args, **kwargs):
        _initialize_mpl()
        super().__init__(*args, **kwargs)
        self.model = model
        # Map Figure UUID to widget with JupyterFigureTab
        self._figures = {}

        for figure_spec in model:
            self._add_figure(figure_spec)
        self.model.events.added.connect(self._on_figure_added)
        self.model.events.removed.connect(self._on_figure_removed)

    @property
    def figures(self):
        "Read-only access to the mapping FigureSpec UUID -> JupyterFigure"
        return DictView(self._figures)

    def _on_figure_added(self, event):
        figure_spec = event.item
        self._add_figure(figure_spec)

    def _add_figure(self, figure_spec):
        "Add a new tab with a matplotlib Figure."
        tab = _JupyterFigureTab(figure_spec, parent=self)
        self._figures[figure_spec.uuid] = tab
        self.children = (*self.children, tab)
        index = len(self.children) - 1
        self.set_title(index, figure_spec.title)
        figure_spec.events.title.connect(self._on_title_changed)
        figure_spec.events.short_title.connect(self._on_short_title_changed)
        # Workaround: If the tabs are cleared and then children are added
        # again, no tab is selected.
        if index == 0:
            self.selected_index = 0

    def _on_figure_removed(self, event):
        "Remove the associated tab and close its canvas."
        figure_spec = event.item
        widget = self._figures[figure_spec.uuid]
        children = list(self.children)
        children.remove(widget)
        self.children = tuple(children)
        widget.close_figure()
        del self._figures[figure_spec.uuid]

    def _on_short_title_changed(self, event):
        "This sets the tab title."
        figure_spec = event.figure_spec
        widget = self._figures[figure_spec.uuid]
        index = self.children.index(widget)
        # Fall back to title if short_title is being unset.
        if event.value is None:
            self.set_title(index, figure_spec.title)
        else:
            self.set_title(index, event.value)

    def _on_title_changed(self, event):
        "This sets the tab title only if short_title is None."
        figure_spec = event.figure_spec
        if figure_spec.short_title is None:
            widget = self._figures[figure_spec.uuid]
            index = self.children.index(widget)
            self.set_title(index, event.value)

    def on_close_tab_requested(self, model):
        # When closing is initiated from the view, remove the associated
        # model.
        self.model.remove(model)


class JupyterFigure(widgets.HBox):
    """
    A Jupyter view for a FigureSpec model. This always contains one Figure.
    """

    def __init__(self, model: FigureSpec):
        _initialize_mpl()
        super().__init__()
        self.model = model
        self.figure, self.axes_list = _make_figure(model)
        self.figure.suptitle(model.title)
        self._axes = {}
        for axes_spec, axes in zip(model.axes, self.axes_list):
            self._axes[axes_spec.uuid] = MatplotlibAxes(model=axes_spec, axes=axes)
        self.children = (self.figure.canvas,)

        model.events.title.connect(self._on_title_changed)

        # By "resizing" (even without actually changing the size) we bump the
        # ipympl machinery that sets up frontend--backend communication and
        # starting displaying data from the figure. Without this, the figure
        # *widget* displays instantly but the actual *plot* (the PNG data sent from
        # matplotlib) is not displayed until cell execution completes.
        _, _, width, height = self.figure.bbox.bounds
        self.figure.canvas.manager.resize(width, height)
        self.figure.canvas.draw_idle()

        # The FigureSpec model does not currently allow axes to be added or
        # removed, so we do not need to handle changes in model.axes.

    @property
    def axes(self):
        "Read-only access to the mapping AxesSpec UUID -> MatplotlibAxes"
        return DictView(self._axes)

    def _on_title_changed(self, event):
        self.figure.suptitle(event.value)
        self._redraw()

    def _redraw(self):
        "Redraw the canvas."
        self.figure.canvas.draw_idle()

    def close_figure(self):
        self.figure.canvas.close()


class _JupyterFigureTab(widgets.HBox):
    """
    A tab in a widgets.Tab container that contains a JupyterFigure.

    This is aware of its parent in order to support tab-closing.
    """

    def __init__(self, model: FigureSpec, parent):
        super().__init__()
        self.model = model
        self.parent = parent
        self.button = widgets.Button(description="Close")
        self.button.on_click(lambda _: self.parent.on_close_tab_requested(self.model))
        self._jupyter_figure = JupyterFigure(model)
        self.children = (self._jupyter_figure, self.button)

        # Pass-through accessors to match the API of QtFigure, which has/needs
        # one less layer.
        self.figure = self._jupyter_figure.figure

    def close_figure(self):
        # Pass through toe JupyterFigure instance.
        return self._jupyter_figure.close_figure()

    @property
    def axes(self):
        "Read-only access to the mapping AxesSpec UUID -> MatplotlibAxes"
        return DictView(self._jupyter_figure.axes)


def _make_figure(figure_spec):
    "Create a Figure and Axes."
    # This import must be deferred until after the matplotlib backend is set,
    # which happens when a JupyterFigure or JupyterFigures is instantiated
    # for the first time.
    import matplotlib.pyplot as plt

    # By default, with interactive mode on, each fig.show() will be called
    # automatically, and we'll get duplicates littering the output area. We
    # only want to see the figures where they are placed explicitly in widgets.
    plt.ioff()

    # TODO Let FigureSpec give different options to subplots here,
    # but verify that number of axes created matches the number of axes
    # specified.
    figure, axes = plt.subplots(len(figure_spec.axes))
    # Handle return type instability in plt.subplots.
    if not isinstance(axes, collections.abc.Iterable):
        axes = [axes]
    return figure, axes



import abc
import collections.abc
from datetime import datetime, timedelta
import itertools

import dateutil.tz

from ..utils.list import EventedList
from ..utils.event import EmitterGroup, Event

LOCAL_TIMEZONE = dateutil.tz.tzlocal()
_epoch = datetime(1970, 1, 1, 0, 0, tzinfo=LOCAL_TIMEZONE)


"""
Vendored from databroker.queries
"""


class InvertedRange(ValueError):
    ...


"""
Vendored from databroker.queries
"""


class Query(collections.abc.Mapping):
    """
    This represents a MongoDB query.

    MongoDB queries are typically encoded as simple dicts. This object supports
    the dict interface in a read-only fashion. Subclassses add a nice __repr__
    and mutable attributes from which the contents of the dict are derived.
    """

    @abc.abstractproperty
    def query(self):
        ...

    @abc.abstractproperty
    def kwargs(self):
        ...

    def __iter__(self):
        return iter(self.query)

    def __getitem__(self, key):
        return self.query[key]

    def __len__(self):
        return len(self.query)

    def replace(self, **kwargs):
        """
        Make a copy with parameters changed.
        """
        return type(self)(**{**self.kwargs, **kwargs})

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            f"{', '.join(f'{k}={v}' for k, v in self.kwargs.items())})"
        )


class TextQuery(Query):
    def __init__(self, text_search):
        self._text_search = text_search

    @property
    def kwargs(self):
        return {
            "text_search": self._text_search,
        }

    @property
    def query(self):
        return {
            "$text": {"$search": self._text_search},
        }


"""
Vendored from databroker.queries
"""


class TimeRange(Query):
    """
    A search query representing a time range.

    Parameters
    ----------
    since, until: dates gives as timestamp, datetime, or human-friendly string, optional
    timezone : string
        As in, 'US/Eastern'. If None is given, tzlocal is used.

    Examples
    --------
    Any granularity (year, month, date, hour, minute, second) is accepted.

    >>> TimeRange(since='2014')

    >>> TimeRange(until='2019-07')

    >>> TimeRange(since='2014-07-04', until='2020-07-04')

    >>> TimeRange(since='2014-07-04 05:00')

    Create a copy replacing some parameter. This leaves the original unchanged.

    >>> tr = TimeRange(since='2014-07-04 05:00')
    >>> tr.replace(until='2015')
    TimeRange(since='2014-07-04 05:00', until='2015')
    >>> tr
    TimeRange(since='2014-07-04 05:00')

    Access the raw query that this generates.

    >>> TimeRange(since='2014').query
    {'time': {'$gte': 1388552400.0}}
    """

    def __init__(self, since=None, until=None, timezone=None):
        if timezone is None:
            import tzlocal

            timezone = tzlocal.get_localzone().zone
        self.timezone = timezone
        if since is None:
            self._since_normalized = None
        else:
            self._since_normalized = normalize_human_friendly_time(
                since, tz=self.timezone
            )
        self._since_raw = since
        if until is None:
            self._until_normalized = None
        else:
            self._until_normalized = normalize_human_friendly_time(
                until, tz=self.timezone
            )
        self._until_raw = until
        if since is not None and until is not None:
            if self._since_normalized > self._until_normalized:
                raise InvertedRange("since must not be greater than until.")

    @property
    def kwargs(self):
        return {
            "since": self._since_raw,
            "until": self._until_raw,
            "timezone": self.timezone,
        }

    @property
    def query(self):
        query = {"time": {}}
        if self._since_normalized is not None:
            query["time"]["$gte"] = self._since_normalized
        if self._until_normalized is not None:
            query["time"]["$lt"] = self._until_normalized
        if query["time"]:
            return query
        else:
            return {}


# The following are vendored from databroker.utils.

# human friendly timestamp formats we'll parse
_TS_FORMATS = [
    "%Y-%m-%d %H:%M:%S",
    "%Y-%m-%d %H:%M",  # these 2 are not as originally doc'd,
    "%Y-%m-%d %H",  # but match previous pandas behavior
    "%Y-%m-%d",
    "%Y-%m",
    "%Y",
]

# build a tab indented, '-' bulleted list of supported formats
# to append to the parsing function docstring below
_doc_ts_formats = "\n".join("\t- {}".format(_) for _ in _TS_FORMATS)


def normalize_human_friendly_time(val, tz):
    """Given one of :
    - string (in one of the formats below)
    - datetime (eg. datetime.now()), with or without tzinfo)
    - timestamp (eg. time.time())
    return a timestamp (seconds since jan 1 1970 UTC).

    Non string/datetime values are returned unaltered.
    Leading/trailing whitespace is stripped.
    Supported formats:
    {}
    """
    # {} is placeholder for formats; filled in after def...

    import pytz

    zone = pytz.timezone(tz)  # tz as datetime.tzinfo object
    epoch = pytz.UTC.localize(datetime(1970, 1, 1))
    check = True

    if isinstance(val, str):
        # unix 'date' cmd format '%a %b %d %H:%M:%S %Z %Y' works but
        # doesn't get TZ?

        # Could cleanup input a bit? remove leading/trailing [ :,-]?
        # Yes, leading/trailing whitespace to match pandas behavior...
        # Actually, pandas doesn't ignore trailing space, it assumes
        # the *current* month/day if they're missing and there's
        # trailing space, or the month is a single, non zero-padded digit.?!
        val = val.strip()

        for fmt in _TS_FORMATS:
            try:
                ts = datetime.strptime(val, fmt)
                break
            except ValueError:
                pass

        try:
            if isinstance(ts, datetime):
                val = ts
                check = False
            else:
                # what else could the type be here?
                raise TypeError("expected datetime," " got {:r}".format(ts))

        except NameError:
            raise ValueError("failed to parse time: " + repr(val))

    if check and not isinstance(val, datetime):
        return val

    if val.tzinfo is None:
        # is_dst=None raises NonExistent and Ambiguous TimeErrors
        # when appropriate, same as pandas
        val = zone.localize(val, is_dst=None)

    return (val - epoch).total_seconds()


# fill in the placeholder we left in the previous docstring
normalize_human_friendly_time.__doc__ = normalize_human_friendly_time.__doc__.format(
    _doc_ts_formats
)


def secs_since_epoch(datetime):
    return (datetime - _epoch) / timedelta(seconds=1)


def ensure_abs(*abs_or_rel_times):
    """
    If datetime, pass through. If timedelta, interpret as relative to now.
    """
    now = datetime.now()
    results = []
    for item in abs_or_rel_times:
        if isinstance(item, timedelta):
            results.append(now + item)
        else:
            results.append(item)
    return results


class SearchInput:
    def __init__(self, *, text_search_supported=False):
        self._since = None
        self._until = None
        self._text = None
        self._query = {}
        self.events = EmitterGroup(
            source=self,
            auto_connect=True,
            query=Event,
            since=Event,
            until=Event,
            reload=Event,
            text=Event,
        )
        self._time_validator = None
        self._text_search_supported = text_search_supported

    @property
    def time_validator(self):
        return self._time_validator

    @time_validator.setter
    def time_validator(self, validator):
        self._time_validator = validator

    @property
    def text_search_supported(self):
        return self._text_search_supported

    def __repr__(self):
        return f"<SearchInput {self._query!r}>"

    @property
    def query(self):
        """
        MongoDB query
        """
        return self._query

    @query.setter
    def query(self, query):
        if query == self.query:
            return
        self._query = query
        self.events.query(query=query)

    @property
    def since(self):
        """
        Lower bound on time range
        """
        return self._since

    @since.setter
    def since(self, since):
        if self.time_validator is not None:
            self.time_validator(since=since, until=self.until)
        if isinstance(since, (int, float)):
            since = datetime.fromtimestamp(since)
        if isinstance(since, datetime):
            if since == self.since:
                return
            if since.tzinfo is None:
                since = since.replace(tzinfo=LOCAL_TIMEZONE)
        self._since = since
        self.events.since(date=since)

    @property
    def until(self):
        """
        Upper bound on time range
        """
        return self._until

    @until.setter
    def until(self, until):
        if self.time_validator is not None:
            self.time_validator(since=self.since, until=until)
        if isinstance(until, (int, float)):
            until = datetime.fromtimestamp(until)
        if isinstance(until, datetime):
            if until == self.until:
                return
            if until.tzinfo is None:
                until = until.replace(tzinfo=LOCAL_TIMEZONE)
        self._until = until
        self.events.until(date=until)

    @property
    def text(self):
        """
        Text search
        """
        return self._text

    @text.setter
    def text(self, text):
        if text and not self.text_search_supported:
            raise RuntimeError("This catalog does not support text search.")
        self._text = text
        self.events.text(text=text)

    def on_text(self, event):
        if not event.text:
            self._query.pop("$text", None)
        else:
            self._query.update(TextQuery(event.text))
        self.events.query(query=self._query)

    def on_since(self, event):
        try:
            since, until = ensure_abs(event.date, self._until)
            tr = TimeRange(since=since, until=until)
        except InvertedRange:
            # Move 'until' as well to create a valid (though empty) interval.
            self.until = event.date
            return
        if tr:
            self._query.update(tr)
        else:
            self._query.pop("time", None)
        self.events.query(query=self._query)

    def on_until(self, event):
        try:
            since, until = ensure_abs(self._since, event.date)
            tr = TimeRange(since=since, until=until)
        except InvertedRange:
            # Move 'since' as well to create a valid (though empty) interval.
            self.since = event.date
            return
        if tr:
            self._query.update(tr)
        else:
            self._query.pop("time", None)
        self.events.query(query=self._query)

    def request_reload(self):
        # If time range was given in relative terms, re-evaluate them relative
        # to the current time.
        changed = False
        if isinstance(self._since, timedelta):
            self.since = self._since
            changed = True
        if isinstance(self._until, timedelta):
            self.until = self._until
            changed = True
        # If the times were re-evaluated, the query Event will have been
        # emitted, so it would be redundant to reload.
        if not changed:
            self.events.reload()


class SearchResults:
    """
    Parameters
    ----------
    columns: tuple
        Expected to have two elements. First is a list of columns names. Second
        is a function that gives the values for one result row.
        Expected signature::
            f(BlueskyRun) -> tuple[str]
    """

    def __init__(self, columns):
        self._catalog = {}
        self._row_cache = {}
        self._selected_rows = EventedList()
        self._active_row = None
        self.columns = columns
        self.events = EmitterGroup(
            source=self,
            auto_connect=True,
            active_row=Event,
            begin_reset=Event,
            end_reset=Event,
        )

    @property
    def headings(self):
        return self._headings

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, columns):
        self._row_cache.clear()
        self._columns = columns
        self._headings, self._row_factory = columns

    @property
    def catalog(self):
        "Catalog of current results"
        return self._catalog

    @catalog.setter
    def catalog(self, catalog):
        self.events.begin_reset()
        self._row_cache.clear()
        self._catalog = catalog
        self._iterator = iter(catalog)
        self._uids = []
        self._selected_rows.clear()
        self.events.end_reset()

    def get_data(self, row, column):
        """
        Get data for one item of the display table.
        """
        uid = self.get_uid_by_row(row)
        # To save on function calls, format one whole row in one step and cache
        # the result.
        try:
            row_content = self._row_cache[uid]
        except KeyError:
            run = self._catalog[uid]
            row_content = self._row_factory(run)
            self._row_cache[uid] = row_content
        item_content = row_content[column]  # content for one cell of the grid
        return item_content

    def get_uid_by_row(self, row):
        if row > len(self._catalog):
            raise ValueError(
                f"Cannot get row {row}. Catalog has {len(self._catalog)} rows."
            )
        cache_length = len(self._uids)
        if row >= cache_length:
            for _ in range(row - cache_length + 1):
                self._uids.append(next(self._iterator))
        return self._uids[row]

    @property
    def active_row(self):
        return self._active_row

    @active_row.setter
    def active_row(self, active_row):
        if active_row != self.active_row:
            self._active_row = active_row
        self.events.active_row(item=active_row)

    @property
    def active_uid(self):
        if self.active_row is not None:
            return self.get_uid_by_row(self.active_row)

    @property
    def active_run(self):
        if self.active_uid is not None:
            return self.catalog[self.active_uid]

    @property
    def selected_rows(self):
        return self._selected_rows

    @property
    def selected_uids(self):
        return [self.get_uid_by_row(row) for row in self._selected_rows]

    @property
    def selection_as_catalog(self):
        "A Catalog containing the selected rows"
        return self.catalog.search({"uid": {"$in": self.selected_uids}})


class RunSearch:
    """
    Model of search input and search results for a search for Runs in a catalog of runs.
    """

    def __init__(self, catalog, columns):
        self.catalog = catalog
        # TODO Choose a gentler way to do this check.
        # The issue here is that only real MongoDB supports $text queries, not
        # the mongoquery library used by JSONL and msgpack databroker drivers
        # or any other "mock" in-memory MongoDB imitators that we know of.
        try:
            catalog.search({"$text": ""})
        except NotImplementedError:
            text_search_supported = False
        else:
            text_search_supported = True
        self.search_input = SearchInput(text_search_supported=text_search_supported)
        self.search_results = SearchResults(columns)
        self.search_input.events.query.connect(self._on_query)
        self.search_input.events.reload.connect(self._on_reload)
        # Initialize the results with the initial state of SearchInput.
        self.search_input.events.query(query=self.search_input.query)

    def _on_reload(self, event):
        self.search_results.events.begin_reset()
        self.search_results.catalog.reload()
        self.search_results.events.end_reset()

    def _on_query(self, event):
        results = self.catalog.search(event.query)
        self.search_results.catalog = results


class Search:
    """
    Model for digging into potentially nested catalogs and ending at a catalog of runs.
    """

    _name_counter = itertools.count(1)

    def __init__(self, root_catalog, *, name=None, columns):
        if name is None:
            name = self.get_default_name()
        self._name = name
        self._subcatalogs = []
        self._root_catalog = root_catalog
        self._columns = columns
        self._search = None
        self._active = False
        self.events = EmitterGroup(
            source=self,
            auto_connect=True,
            enter=Event,
            go_back=Event,
            run_search_ready=Event,
            run_search_cleared=Event,
            active=Event,
            inactive=Event,
            active_run=Event,
        )

        if self._has_runs(root_catalog):
            self._search = RunSearch(root_catalog, columns)
            self._search.search_results.events.active_row.connect(self._on_active_row)
            self.events.run_search_ready(
                search_input=self._search.search_input,
                search_results=self._search.search_results,
            )

    @classmethod
    def get_default_name(cls):
        return next(cls._name_counter)

    @property
    def name(self):
        return self._name

    @property
    def run_search(self):
        return self._search

    @property
    def root_catalog(self):
        return self._root_catalog

    @property
    def current_catalog(self):
        if self._subcatalogs:
            return self._subcatalogs[-1]
        else:
            return self._root_catalog

    @property
    def breadcrumbs(self):
        "Names of subcatalogs tranversed"
        return [catalog.name for catalog in self._subcatalogs]

    @property
    def input(self):
        # a convenience accessor
        if self._search is not None:
            return self._search.search_input

    @property
    def results(self):
        # For convenience in console, having in mind usage like
        # >>> viewer.active_search.results
        # to get the catalog.
        if self._search is not None:
            return self._search.search_results.catalog

    @property
    def selected_uids(self):
        if self._search is not None:
            return self._search.search_results.selected_uids

    @property
    def selection_as_catalog(self):
        if self._search is not None:
            return self._search.search_results.selection_as_catalog

    def enter(self, name):
        if self._search is not None:
            raise RuntimeError("Already all the way into a Catalog of Runs")
        if self._subcatalogs:
            old = self._subcatalogs[-1]
        else:
            old = self._root_catalog
        new = old[name]

        # Touch an attribute that will trigger a connection attempt. (It's here
        # that an error would be raised if, say, a database is unreachable.)
        new.metadata

        # If we get this far, it worked.
        self._subcatalogs.append(new)
        if not self._has_runs(new):
            # Step through another subcatalog.
            self.events.enter(catalog=new)
        else:
            self._search = RunSearch(new, self._columns)
            self._search.search_results.events.active_row.connect(self._on_active_row)
            self.events.run_search_ready(
                search_input=self._search.search_input,
                search_results=self._search.search_results,
            )

    def go_back(self):
        if self.root_catalog is self.current_catalog:
            raise RuntimeError("We are the root catalog.")
        if self._search is not None:
            self._search.search_results.events.active_row.disconnect(
                self._on_active_row
            )
            self._search = None
            self.events.run_search_cleared()
        self._subcatalogs.pop()
        self.events.go_back()

    def _on_active_row(self, event):
        self.events.active_run(uid=self.active_uid, run=self.active_run)

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, active):
        if active == self.active:
            return
        self._active = active
        if active:
            self.events.active()
        else:
            self.events.inactive()

    @property
    def active_uid(self):
        if self._search is not None:
            return self._search.search_results.active_uid

    @property
    def active_run(self):
        if self._search is not None:
            return self._search.search_results.active_run

    @staticmethod
    def _has_runs(catalog):
        "Is this a catalog BlueskyRuns, or a Catalog of Catalogs?"
        # HACK!
        from databroker.v2 import Broker

        return isinstance(catalog, Broker)


class SearchList(EventedList):
    """
    Model for a list of Search models
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.events.add(active=Event)
        self.events.added.connect(self._connect_enforce_mutually_exclusive_activation)
        self.events.removed.connect(
            self._disconnect_enforce_mutually_exclusive_activation
        )

    # Ensure that whenever an item in this list become "active" all others are
    # not active.

    def _connect_enforce_mutually_exclusive_activation(self, event):
        event.item.events.active.connect(self._enforce_mutually_exclusive_activation)

    def _disconnect_enforce_mutually_exclusive_activation(self, event):
        event.item.events.active.disconnect(self._enforce_mutually_exclusive_activation)

    def _enforce_mutually_exclusive_activation(self, event):
        for item in self:
            if item is not event.source:
                item.active = False
        self.events.active(item=event.source)

    @property
    def active(self):
        "The active item in the list, if any"
        for item in self:
            if item.active:
                return item



import pickle

from bluesky.run_engine import Dispatcher, DocumentNames
import zmq

from ..qt.threading import create_worker
from qtpy.QtCore import QTimer, QObject


LOADING_LATENCY = 0.01


class RemoteDispatcher(QObject):
    """
    Dispatch documents received over the network from a 0MQ proxy.

    This is designed to be run in a Qt application.

    Parameters
    ----------
    address : tuple | str
        Address of a running 0MQ proxy, given either as a string like
        ``'127.0.0.1:5567'`` or as a tuple like ``('127.0.0.1', 5567)``
    prefix : bytes, optional
        User-defined bytestring used to distinguish between multiple
        Publishers. If set, messages without this prefix will be ignored.
        If unset, no mesages will be ignored.
    deserializer: function, optional
        optional function to deserialize data. Default is pickle.loads

    Examples
    --------

    Print all documents generated by remote RunEngines.

    >>> d = RemoteDispatcher(('localhost', 5568))
    >>> d.subscribe(stream_documents_into_runs(model.add_run))
    >>> d.start()  # launches periodic workers on background threads
    >>> d.stop()  # stop launching workers
    """

    def __init__(self, address, *, prefix=b"", deserializer=pickle.loads, parent=None):
        super().__init__(parent)
        if isinstance(prefix, str):
            raise ValueError("prefix must be bytes, not string")
        if b" " in prefix:
            raise ValueError("prefix {!r} may not contain b' '".format(prefix))
        self._prefix = prefix
        if isinstance(address, str):
            address = address.split(":", maxsplit=1)
        self._deserializer = deserializer
        self.address = (address[0], int(address[1]))

        self._context = zmq.Context()
        self._socket = self._context.socket(zmq.SUB)
        url = "tcp://%s:%d" % self.address
        self._socket.connect(url)
        self._socket.setsockopt_string(zmq.SUBSCRIBE, "")
        self._task = None
        self.closed = False
        self._timer = QTimer(self)
        self._dispatcher = Dispatcher()
        self.subscribe = self._dispatcher.subscribe
        self._waiting_for_start = True

    def _receive_data(self):
        our_prefix = self._prefix  # local var to save an attribute lookup
        # TODO Pull on the socket more than once here, until it blocks, to
        # ensure we do not get more and more behind over time.
        message = self._socket.recv()
        prefix, name, doc = message.split(b" ", 2)
        name = name.decode()
        if (not our_prefix) or prefix == our_prefix:
            if self._waiting_for_start:
                if name == "start":
                    self._waiting_for_start = False
                else:
                    # We subscribed midstream and are seeing documents for
                    # which we do not have the full run. Wait for a 'start'
                    # doc.
                    return
            doc = self._deserializer(doc)
        return name, doc

    def start(self):
        if self.closed:
            raise RuntimeError(
                "This RemoteDispatcher has already been "
                "started and interrupted. Create a fresh "
                "instance with {}".format(repr(self))
            )
        self._work_loop()

    def _work_loop(self):
        worker = create_worker(
            self._receive_data,
        )
        # Schedule this method to be run again after a brief wait.
        worker.finished.connect(
            lambda: self._timer.singleShot(LOADING_LATENCY, self._work_loop)
        )
        worker.returned.connect(self._process_result)
        worker.start()

    def _process_result(self, result):
        if result is None:
            return
        name, doc = result
        self._dispatcher.process(DocumentNames[name], doc)

    def stop(self):
        self.closed = True



import functools
import itertools

import numpy

from .plot_specs import (
    FigureSpec,
    AxesSpec,
    ImageSpec,
    LineSpec,
)
from .utils import auto_label, call_or_eval, RunManager, run_is_live_and_not_completed
from ..utils.dict_view import DictView


class Lines:
    """
    Plot ys vs x for the last N runs.

    This supports plotting columns like ``"I0"`` but also Python
    expressions like ``"5 * log(I0/It)"`` and even
    ``"my_custom_function(I0)"``. See examples below. Consult
    :func:`bluesky_widgets.models.utils.construct_namespace` for details
    about the available variables.

    Parameters
    ----------
    x : String | Callable
        Field name (e.g. "theta") or expression (e.g. "- deg2rad(theta) / 2")
        or callable with expected signature::

            f(run: BlueskyRun) -> x: Array

        Other signatures are also supported to allow for a somewhat "magical"
        usage. See examples below, and also see
        :func:`bluesky_widgets.models.utils.call_or_eval` for details and more
        examples.

    ys : List[String | Callable]
        Field name (e.g. "theta") or expression (e.g. "- deg2rad(theta) / 2")
        or callable with expected signature::

            f(run: BlueskyRun) -> y: Array

        Other signatures are also supported to allow for a somewhat "magical"
        usage. See examples below, and also see
        :func:`bluesky_widgets.models.utils.call_or_eval` for details and more
        examples.

    max_runs : Integer
        Number of Runs to visualize at once. Default is 10.

    label_maker : Callable, optional
        Expected signature::

            f(run: BlueskyRun, y: String) -> label: String

    needs_streams : List[String], optional
        Streams referred to by x and y. Default is ``["primary"]``
    namespace : Dict, optional
        Inject additional tokens to be used in expressions for x and y
    axes : AxesSpec, optional
        If None, an axes and figure are created with default labels and titles
        derived from the ``x`` and ``y`` parameters.

    Attributes
    ----------
    max_runs : int
        Number of Runs to visualize at once. This may be changed at any point.
        (Note: Increasing it will not restore any Runs that have already been
        removed, but it will allow more new Runs to be added.) Runs added
        with ``pinned=True`` are exempt from the limit.
    runs : RunList[BlueskyRun]
        As runs are appended entries will be removed from the beginning of the
        last (first in, first out) so that there are at most ``max_runs``.
    pinned : Frozenset[String]
        Run uids of pinned runs.
    figure : FigureSpec
    axes : AxesSpec
    x : String | Callable
        Read-only access to x
    ys : Tuple[String | Callable]
        Read-only access to ys
    needs_streams : Tuple[String]
        Read-only access to stream names needed
    namespace : Dict
        Read-only access to user-provided namespace

    Examples
    --------

    Plot "det" vs "motor" and view it.

    >>> model = Lines("motor", ["det"])
    >>> from bluesky_widgets.jupyter.figures import JupyterFigure
    >>> view = JupyterFigure(model.figure)
    >>> model.add_run(run)
    >>> model.add_run(another_run, pinned=True)

    Plot a mathematical transformation of the columns using any object in
    numpy. This can be given as a string expression:

    >>> model = Lines("abs(motor)", ["-log(det)"])
    >>> model = Lines("abs(motor)", ["pi * det"])
    >>> model = Lines("abs(motor)", ["sqrt(det)"])

    Plot multiple lines.

    >>> model = Lines("motor", ["log(I0/It)", "log(I0)", "log(It)"])

    Plot every tenth point.

    >>> model = Lines("motor", ["intesnity[::10]"])

    Access data outside the "primary" stream, such as a stream name "baseline".

    >>> model = Lines("motor", ["intensity/baseline['intensity'][0]"])

    As shown, objects from numpy can be used in expressions. You may define
    additional words, such as "savlog" for a Savitzky-Golay smoothing filter,
    by passing it a dict mapping the new word to the new object.

    >>> import scipy.signal
    >>> namespace = {"savgol": scipy.signal.savgol_filter}
    >>> model = Lines("motor", ["savgol(intensity, 5, 2)"],
    ...                     namespace=namespace)

    Or you may pass in a function. It will be passed parameters according to
    their names.

    >>> model = Lines("motor", [lambda intensity: savgol(intensity, 5, 2)])

    More examples of this function-based usage:

    >>> model = Lines("abs(motor)", [lambda det: -log(det)])
    >>> model = Lines("abs(motor)", [lambda det, pi: pi * det])
    >>> model = Lines("abs(motor)", [lambda det, np: np.sqrt(det)])

    Custom, user-defined objects may be added in the same way, either by adding
    names to the namespace or providing the functions directly.
    """

    def __init__(
        self,
        x,
        ys,
        *,
        max_runs=10,
        label_maker=None,
        needs_streams=("primary",),
        namespace=None,
        axes=None,
    ):
        super().__init__()

        if label_maker is None:
            # scan_id is always generated by RunEngine but not stricter required by
            # the schema, so we fail gracefully if it is missing.

            if len(ys) > 1:

                def label_maker(run, y):
                    return (
                        f"Scan {run.metadata['start'].get('scan_id', '?')} "
                        f"{auto_label(y)}"
                    )

            else:

                def label_maker(run, y):
                    return f"Scan {run.metadata['start'].get('scan_id', '?')}"

        self._x = x
        if isinstance(ys, str):
            raise ValueError("`ys` must be a list of strings, not a string")
        self._ys = tuple(ys)
        self._label_maker = label_maker
        self._namespace = namespace
        if axes is None:
            axes = AxesSpec(
                x_label=auto_label(self.x),
                y_label=", ".join(auto_label(y) for y in self.ys),
            )
            figure = FigureSpec((axes,), title=f"{axes.y_label} v {axes.x_label}")
        else:
            figure = axes.figure
        self.axes = axes
        self.figure = figure
        # If the Axes' figure is not yet set, listen for it to be set.
        if figure is None:

            def set_figure(event):
                self.figure = event.value
                # This occurs at most once, so we can now stop listening.
                self.axes.events.figure.disconnect(set_figure)

            self.axes.events.figure.connect(set_figure)

        self._color_cycle = itertools.cycle(f"C{i}" for i in range(10))

        self._run_manager = RunManager(max_runs, needs_streams)
        self._run_manager.events.run_ready.connect(self._add_lines)
        self.add_run = self._run_manager.add_run
        self.discard_run = self._run_manager.discard_run

    def _transform(self, run, x, y):
        return call_or_eval((x, y), run, self.needs_streams, self.namespace)

    def _add_lines(self, event):
        "Add a line."
        run = event.run
        for y in self.ys:
            label = self._label_maker(run, y)
            # If run is in progress, give it a special color so it stands out.
            if run_is_live_and_not_completed(run):
                color = "black"

                def restyle_line_when_complete(event):
                    "When run is complete, update style."
                    line.style.update({"color": next(self._color_cycle)})

                run.events.completed.connect(restyle_line_when_complete)
            else:
                color = next(self._color_cycle)
            style = {"color": color}

            # Style pinned runs differently.
            if run.metadata["start"]["uid"] in self.pinned:
                style.update(linestyle="dashed")
                label += " (pinned)"

            func = functools.partial(self._transform, x=self.x, y=y)
            line = LineSpec(func, run, label, style)
            self._run_manager.track_artist(line)
            self.axes.lines.append(line)

    @property
    def x(self):
        return self._x

    @property
    def ys(self):
        return self._ys

    @property
    def namespace(self):
        return DictView(self._namespace or {})

    # Expose some properties from the internal RunManger helper class.

    @property
    def runs(self):
        return self._run_manager.runs

    @property
    def max_runs(self):
        return self._run_manager.max_runs

    @max_runs.setter
    def max_runs(self, value):
        self._run_manager.max_runs = value

    @property
    def needs_streams(self):
        return self._run_manager._needs_streams

    @property
    def pinned(self):
        return self._run_manager._pinned


class Images:
    """
    Plot an image from a Run.

    By default, higher-dimensional data is handled by repeatedly averaging over
    the leading dimension until there are only two dimensions.

    Parameters
    ----------

    field : string
        Field name or expression
    max_runs : Integer
        Number of Runs to visualize at once. Default is 1.
    label_maker : Callable, optional
        Expected signature::

            f(run: BlueskyRun, y: String) -> label: String

    needs_streams : List[String], optional
        Streams referred to by field. Default is ``["primary"]``
    namespace : Dict, optional
        Inject additional tokens to be used in expressions for x and y
    axes : AxesSpec, optional
        If None, an axes and figure are created with default labels and titles
        derived from the ``x`` and ``y`` parameters.

    Attributes
    ----------
    max_runs : int
        Number of Runs to visualize at once. This may be changed at any point.
        (Note: Increasing it will not restore any Runs that have already been
        removed, but it will allow more new Runs to be added.) Runs added
        with ``pinned=True`` are exempt from the limit.
    runs : RunList[BlueskyRun]
        As runs are appended entries will be removed from the beginning of the
        last (first in, first out) so that there are at most ``max_runs``.
    pinned : Frozenset[String]
        Run uids of pinned runs.
    figure : FigureSpec
    axes : AxesSpec
    field : String
        Read-only access to field or expression
    needs_streams : List[String], optional
        Read-only access to streams referred to by field.
    namespace : Dict, optional
        Read-only access to user-provided namespace

    Examples
    --------
    >>> model = Images("ccd")
    >>> from bluesky_widgets.jupyter.figures import JupyterFigure
    >>> view = JupyterFigure(model.figure)
    >>> model.add_run(run)
    """

    # TODO: fix x and y limits here

    def __init__(
        self,
        field,
        *,
        max_runs=1,
        label_maker=None,
        needs_streams=("primary",),
        namespace=None,
        axes=None,
    ):
        super().__init__()

        if label_maker is None:
            # scan_id is always generated by RunEngine but not stricter required by
            # the schema, so we fail gracefully if it is missing.

            def label_maker(run, field):
                md = run.metadata["start"]
                return (
                    f"Scan ID {md.get('scan_id', '?')}   UID {md['uid'][:8]}   "
                    f"{auto_label(field)}"
                )

        self._field = field
        self._label_maker = label_maker
        self._namespace = namespace
        if axes is None:
            axes = AxesSpec()
            figure = FigureSpec((axes,), title="")
        else:
            figure = axes.figure
        self.axes = axes
        self.figure = figure
        # If the Axes' figure is not yet set, listen for it to be set.
        if figure is None:

            def set_figure(event):
                self.figure = event.value
                # This occurs at most once, so we can now stop listening.
                self.axes.events.figure.disconnect(set_figure)

            self.axes.events.figure.connect(set_figure)

        self._run_manager = RunManager(max_runs, needs_streams)
        self._run_manager.events.run_ready.connect(self._add_images)
        self.add_run = self._run_manager.add_run
        self.discard_run = self._run_manager.discard_run

    def _add_images(self, event):
        run = event.run
        func = functools.partial(self._transform, field=self.field)
        image = ImageSpec(func, run, label=self.field)
        array_shape = run.primary.read()[self.field].shape
        self._run_manager.track_artist(image)
        self.axes.images.append(image)
        self.axes.title = self._label_maker(run, self.field)
        # By default, pixels center on integer coordinates ranging from 0 to
        # columns-1 horizontally and 0 to rows-1 vertically.
        # In order to see entire pixels, we set lower limits to -0.5
        # and upper limits to columns-0.5 horizontally and rows-0.5 vertically
        # if limits aren't specifically set.
        if self.axes.x_limits is None:
            self.axes.x_limits = (-0.5, array_shape[-1] - 0.5)
        if self.axes.y_limits is None:
            self.axes.y_limits = (-0.5, array_shape[-2] - 0.5)
        # TODO Set axes x, y from xarray dims

    def _transform(self, run, field):
        (data,) = numpy.asarray(
            call_or_eval((field,), run, self.needs_streams, self.namespace)
        )
        # If the data is more than 2D, take the middle slice from the leading
        # axis until there are only two axes.
        while data.ndim > 2:
            middle = data.shape[0] // 2
            data = data[middle]
        return data

    @property
    def field(self):
        return self._field

    @property
    def namespace(self):
        return DictView(self._namespace or {})

    # Expose some properties from the internal RunManger helper class.

    @property
    def runs(self):
        return self._run_manager.runs

    @property
    def max_runs(self):
        return self._run_manager.max_runs

    @max_runs.setter
    def max_runs(self, value):
        self._run_manager.max_runs = value

    @property
    def needs_streams(self):
        return self._run_manager._needs_streams

    @property
    def pinned(self):
        return self._run_manager._pinned


class RasteredImages:
    """
    Plot a rastered image from a Run.

    Parameters
    ----------

    field : string
        Field name or expression
    shape : Tuple[Integer]
        The (row, col) shape of the raster
    label_maker : Callable, optional
        Expected signature::

            f(run: BlueskyRun, y: String) -> label: String

    needs_streams : List[String], optional
        Streams referred to by field. Default is ``["primary"]``
    namespace : Dict, optional
        Inject additional tokens to be used in expressions for x and y
    axes : AxesSpec, optional
        If None, an axes and figure are created with default labels and titles
        derived from the ``x`` and ``y`` parameters.
    clim : Tuple, optional
        The color limits
    cmap : String or Colormap, optional
        The color map to use
    extent : scalars (left, right, bottom, top), optional
        Passed through to :meth:`matplotlib.axes.Axes.imshow`
    x_positive : String, optional
        Defines the positive direction of the x axis, takes the values 'right'
        (default) or 'left'.
    y_positive : String, optional
        Defines the positive direction of the y axis, takes the values 'up'
        (default) or 'down'.

    Attributes
    ----------
    run : BlueskyRun
        The currently-viewed Run
    figure : FigureSpec
    axes : AxesSpec
    field : String
        Read-only access to field or expression
    needs_streams : List[String], optional
        Read-only access to streams referred to by field.
    namespace : Dict, optional
        Read-only access to user-provided namespace

    Examples
    --------
    >>> model = RasteredImages("intensity", shape=(100, 200))
    >>> from bluesky_widgets.jupyter.figures import JupyterFigure
    >>> view = JupyterFigure(model.figure)
    >>> model.add_run(run)
    """

    def __init__(
        self,
        field,
        shape,
        *,
        max_runs=1,
        label_maker=None,
        needs_streams=("primary",),
        namespace=None,
        axes=None,
        clim=None,
        cmap="viridis",
        extent=None,
        x_positive="right",
        y_positive="up",
    ):
        super().__init__()

        if label_maker is None:
            # scan_id is always generated by RunEngine but not stricter required by
            # the schema, so we fail gracefully if it is missing.

            def label_maker(run, field):
                md = run.metadata["start"]
                return (
                    f"Scan ID {md.get('scan_id', '?')}   UID {md['uid'][:8]}   {field}"
                )

        self._label_maker = label_maker

        # Stash these and expose them as read-only properties.
        self._field = field
        self._shape = shape
        self._namespace = namespace

        self._run = None

        if axes is None:
            axes = AxesSpec()
            figure = FigureSpec((axes,), title="")
        else:
            figure = axes.figure
        self.axes = axes
        self.figure = figure
        # If the Axes' figure is not yet set, listen for it to be set.
        if figure is None:

            def set_figure(event):
                self.figure = event.value
                # This occurs at most once, so we can now stop listening.
                self.axes.events.figure.disconnect(set_figure)

            self.axes.events.figure.connect(set_figure)
        self._clim = clim
        self._cmap = cmap
        self._extent = extent
        self._x_positive = x_positive
        self._y_positive = y_positive

        self._run_manager = RunManager(max_runs, needs_streams)
        self._run_manager.events.run_ready.connect(self._add_image)
        self.add_run = self._run_manager.add_run
        self.discard_run = self._run_manager.discard_run

    @property
    def cmap(self):
        return self._cmap

    @cmap.setter
    def cmap(self, value):
        self._cmap = value
        for image in self.axes.images:
            image.style.update({"cmap": value})

    @property
    def clim(self):
        return self._clim

    @clim.setter
    def clim(self, value):
        self._clim = value
        for image in self.axes.images:
            image.style.update({"clim": value})

    @property
    def extent(self):
        return self._extent

    @extent.setter
    def extent(self, value):
        self._extent = value
        for image in self.axes.images:
            image.style.update({"extent": value})

    @property
    def x_positive(self):
        xmin, xmax = self.axes.x_limits
        if xmin > xmax:
            self._x_positive = "left"
        else:
            self._x_positive = "right"
        return self._x_positive

    @x_positive.setter
    def x_positive(self, value):
        if value not in ["right", "left"]:
            raise ValueError('x_positive must be "right" or "left"')
        self._x_positive = value
        xmin, xmax = self.axes.x_limits
        if (xmin > xmax and self._x_positive == "right") or (
            xmax > xmin and self._x_positive == "left"
        ):
            self.axes.x_limits = (xmax, xmin)
        elif (xmax >= xmin and self._x_positive == "right") or (
            xmin >= xmax and self._x_positive == "left"
        ):
            self.axes.x_limits = (xmin, xmax)
            self._x_positive = value

    @property
    def y_positive(self):
        ymin, ymax = self.axes.y_limits
        if ymin > ymax:
            self._y_positive = "down"
        else:
            self._y_positive = "up"
        return self._y_positive

    @y_positive.setter
    def y_positive(self, value):
        if value not in ["up", "down"]:
            raise ValueError('y_positive must be "up" or "down"')
        self._y_positive = value
        ymin, ymax = self.axes.y_limits
        if (ymin > ymax and self._y_positive == "up") or (
            ymax > ymin and self._y_positive == "down"
        ):
            self.axes.y_limits = (ymax, ymin)
        elif (ymax >= ymin and self._y_positive == "up") or (
            ymin >= ymax and self._y_positive == "down"
        ):
            self.axes.y_limits = (ymin, ymax)
            self._y_positive = value

    def _add_image(self, event):
        run = event.run
        func = functools.partial(self._transform, field=self.field)
        style = {"cmap": self._cmap, "clim": self._clim, "extent": self._extent}
        image = ImageSpec(func, run, label=self.field, style=style)
        self._run_manager.track_artist(image)
        md = run.metadata["start"]
        self.axes.images.append(image)
        self.axes.title = self._label_maker(run, self.field)
        self.axes.x_label = md["motors"][1]
        self.axes.y_label = md["motors"][0]
        # By default, pixels center on integer coordinates ranging from 0 to
        # columns-1 horizontally and 0 to rows-1 vertically.
        # In order to see entire pixels, we set lower limits to -0.5
        # and upper limits to columns-0.5 horizontally and rows-0.5 vertically
        # if limits aren't specifically set.
        if self.axes.x_limits is None and self._x_positive == "right":
            self.axes.x_limits = (-0.5, md["shape"][1] - 0.5)
        elif self.axes.x_limits is None and self._x_positive == "left":
            self.axes.x_limits = (md["shape"][1] - 0.5, -0.5)
        if self.axes.y_limits is None and self._y_positive == "up":
            self.axes.y_limits = (-0.5, md["shape"][0] - 0.5)
        elif self.axes.y_limits is None and self._y_positive == "down":
            self.axes.y_limits = (md["shape"][0] - 0.5, -0.5)
        # TODO Try to make the axes aspect equal unless the extent is highly non-square.
        ...

    def _transform(self, run, field):
        image_data = numpy.ones(self._shape) * numpy.nan
        (data,) = numpy.asarray(
            call_or_eval((field,), run, self.needs_streams, self.namespace)
        )
        snaking = run.metadata["start"]["snaking"]
        for index in range(len(data)):
            pos = list(numpy.unravel_index(index, self._shape))
            if snaking[1] and (pos[0] % 2):
                pos[1] = self._shape[1] - pos[1] - 1
            pos = tuple(pos)
            image_data[pos] = data[index]

        return image_data

    @property
    def namespace(self):
        return DictView(self._namespace or {})

    @property
    def field(self):
        return self._field

    @property
    def shape(self):
        return self._shape

    # Expose some properties from the internal RunManger helper class.

    @property
    def runs(self):
        return self._run_manager.runs

    @property
    def max_runs(self):
        return self._run_manager.max_runs

    @max_runs.setter
    def max_runs(self, value):
        self._run_manager.max_runs = value

    @property
    def needs_streams(self):
        return self._run_manager._needs_streams

    @property
    def pinned(self):
        return self._run_manager._pinned



import asyncio
import logging
from multiprocessing import Process, Queue
from pathlib import Path
import tempfile

from bluesky.callbacks.zmq import Publisher
from bluesky.callbacks import LiveTable
from bluesky.preprocessors import SupplementalData
from suitcase.jsonl import Serializer
from bluesky import RunEngine
from ophyd.sim import det, motor, motor1, motor2
from bluesky.plans import scan
from event_model import RunRouter
from bluesky.plan_stubs import sleep


log = logging.getLogger(__name__)


def run_proxy(queue):
    """
    Run Proxy on random, free ports and communicate the port numbers back.
    """
    from bluesky.callbacks.zmq import Proxy

    proxy = Proxy()
    queue.put((proxy.in_port, proxy.out_port))
    proxy.start()


def run_publisher(in_port, data_path, quiet=False):
    """
    Acquire data in an infinite loop and publish it.
    """
    publisher = Publisher(f"localhost:{in_port}")
    RE = RunEngine(loop=asyncio.new_event_loop())
    sd = SupplementalData()
    RE.preprocessors.append(sd)
    sd.baseline.extend([motor1, motor2])
    RE.subscribe(publisher)

    def factory(name, doc):
        serializer = Serializer(data_path / "abc", flush=True)
        return [serializer], []

    rr = RunRouter([factory])
    RE.subscribe(rr)
    if not quiet:
        RE.subscribe(LiveTable(["motor", "det"]))

    motor.delay = 0.2
    det.kind = "hinted"

    def infinite_plan():
        while True:
            for i in range(1, 5):
                yield from sleep(2)
                yield from scan([det], motor, -1, 1, 5 * i)

    # Just as a convenience, avoid collission with scan_ids of runs in Catalog.
    RE.md["scan_id"] = 100
    try:
        RE(infinite_plan())
    finally:
        RE.halt()


def stream_example_data(data_path, quiet=False):
    data_path = Path(data_path)
    log.info(
        f"Writing example data into directory {data_path!s}. "
        "It will be deleted when this process is stopped."
    )

    queue = Queue()
    proxy_process = Process(target=run_proxy, args=(queue,))
    proxy_process.start()
    in_port, out_port = queue.get()
    log.info(f"Connect a consumer to localhost:{out_port}")

    publisher_process = Process(target=run_publisher, args=(in_port, data_path, quiet))
    publisher_process.start()
    log.info("Demo acquisition has started.")

    return f"localhost:{out_port}", proxy_process, publisher_process


if __name__ == "__main__":
    import logging
    import sys

    quiet = (len(sys.argv) > 1) and (sys.argv[1] in ("--quiet", "-q"))

    handler = logging.StreamHandler()
    handler.setLevel("INFO")
    log.setLevel("DEBUG")
    log.addHandler(handler)
    with tempfile.TemporaryDirectory() as directory:
        stream_example_data(directory, quiet)
    # Delete example data at exit.



"""
Displays a BlueskyRun in a QTreeView lazily populating the view. In a real
application this would be used in conjunction with a data browser/search box or
on the command line to explore runs.
"""
from bluesky_widgets.models.run_tree import RunTree
from bluesky_widgets.qt import Window
from bluesky_widgets.qt import gui_qt
from bluesky_widgets.qt.run_tree import QtTreeView
from bluesky_widgets.examples.utils.generate_msgpack_data import get_catalog


class RunTree(RunTree):
    """
    A user-facing model extended with a Qt widget and window.
    """

    def __init__(self, run=None, *, show=True, title="Bluesky run tree"):
        super().__init__(run=run)
        self.title = title
        widget = QtTreeView(self)
        self._window = Window(widget, show=show)

    def show(self):
        """Resize, show, and raise the window."""
        self._window.show()

    def close(self):
        """Close the window."""
        self._window.close()


def main():
    print(__doc__)
    with gui_qt("Example Application"):
        tree = RunTree(title="Example Application")

        # You can set your own catalog/run here or use this synthetic data.
        run = get_catalog()[-1]
        tree.run = run


if __name__ == "__main__":
    main()



import collections.abc
from pathlib import Path
import matplotlib

from ..models.plot_specs import FigureSpec, FigureSpecList
from .._matplotlib_axes import MatplotlibAxes
from ..utils.dict_view import DictView


class HeadlessFigures:
    """
    A headless "view" for a FigureSpecList model.

    It does not produce a graphical user interface. Instead, it provides
    methods for exporting figures as images.

    Examples
    --------

    Export all the figures to a directory. They will be named by their title.
    If there are duplciate titles, a counting number will appended like
    x-1.png, x-2.png.

    >>> headless = HeadlessFigures(model)
    >>> headless.export_all("path/to/directory/")

    Control the format.

    >>> headless.export_all("path/to/directory/", format="png")
    >>> headless.export_all("path/to/directory/", format="jpg")
    """

    def __init__(self, model: FigureSpecList):

        self.model = model
        # Map Figure UUID to widget with HeadlessFigure
        self._figures = {}

        for figure_spec in model:
            self._add_figure(figure_spec)
        model.events.added.connect(self._on_figure_added)
        model.events.removed.connect(self._on_figure_removed)

    @property
    def figures(self):
        "Read-only access to the mapping FigureSpec UUID -> HeadlessFigure"
        return DictView(self._figures)

    def _on_figure_added(self, event):
        figure_spec = event.item
        self._add_figure(figure_spec)

    def _add_figure(self, figure_spec):
        "Create a new matplotlib Figure."
        figure = HeadlessFigure(figure_spec)
        self._figures[figure_spec.uuid] = figure

    def _on_figure_removed(self, event):
        "Remove the associated tab and close its canvas."
        figure_spec = event.item
        figure = self._figures[figure_spec.uuid]
        figure.close_figure()
        del self._figures[figure_spec.uuid]

    def close_figures(self):
        for figure in self._figures.values():
            figure.close_figure()

    close = close_figures

    def export_all(self, directory, format="png", **kwargs):
        """
        Export all figures.

        Parameters
        ----------
        directory : str | Path
        format : str, optional
            Default is "png".
        **kwargs :
            Passed through to matplotlib.figure.Figure.savefig

        Returns
        -------
        filenames : List[String]
        """
        # Avoid name collisions in the case of duplicate titles by appending
        # "-1", "-2", "-3", ... to duplicates.
        titles_tallied = {}
        filenames = []
        for figure_spec in self.model:
            title = figure_spec.title
            if title in titles_tallied:
                filename = f"{title}-{titles_tallied[title]}"
                titles_tallied[title] += 1
            else:
                filename = title
                titles_tallied[title] = 1
            filename = str(Path(directory, f"{filename}.{format}"))
            figure = self._figures[figure_spec.uuid]
            figure.export(filename, format=format, **kwargs)
            filenames.append(filename)
        return filenames


class HeadlessFigure:
    """
    A Headless "view" for a FigureSpec model. This always contains one Figure.

    Examples
    --------

    Export the figure.

    >>> headless = HeadlessFigure(model)
    >>> headless.export("my-figure.png")
    """

    def __init__(self, model: FigureSpec):
        self.model = model
        self.figure, self.axes_list = _make_figure(model)
        self.figure.suptitle(model.title)
        self._axes = {}
        for axes_spec, axes in zip(model.axes, self.axes_list):
            self._axes[axes_spec.uuid] = MatplotlibAxes(model=axes_spec, axes=axes)

        model.events.title.connect(self._on_title_changed)
        # The FigureSpec model does not currently allow axes to be added or
        # removed, so we do not need to handle changes in model.axes.

    @property
    def axes(self):
        "Read-only access to the mapping AxesSpec UUID -> MatplotlibAxes"
        return DictView(self._axes)

    def _on_title_changed(self, event):
        self.figure.suptitle(event.value)

    def close_figure(self):
        _close_figure(self.figure)

    close = close_figure

    def export(self, filename, format="png", **kwargs):
        """
        Export figure.

        Parameters
        ----------
        filename : str | Path
        format : str, optional
            Default is "png".
        **kwargs :
            Passed through to matplotlib.figure.Figure.savefig
        """
        self.figure.savefig(str(filename), format=format, **kwargs)


def _make_figure(figure_spec):
    "Create a Figure and Axes."
    matplotlib.use("Agg")  # must set before importing matplotlib.pyplot
    import matplotlib.pyplot as plt  # noqa

    # TODO Let FigureSpec give different options to subplots here,
    # but verify that number of axes created matches the number of axes
    # specified.
    fig, axes = plt.subplots(len(figure_spec.axes))
    # Handl return type instability in plt.subplots.
    if not isinstance(axes, collections.abc.Iterable):
        axes = [axes]
    return fig, axes


def _close_figure(figure):
    """
    Workaround for matplotlib regression relating to closing figures in Agg

    See https://github.com/matplotlib/matplotlib/pull/18184/
    """
    # TODO It would be better to switch the approach based on matplotlib
    # versions known to have this problem, rather than blindly trying. Update
    # this once a fixed has been released and we know the earliest version of
    # matplotlib that does not have this bug.
    try:
        figure.canvas.close()
    except AttributeError:
        from matplotlib._pylab_helpers import Gcf

        num = next(
            (
                manager.num
                for manager in Gcf.figs.values()
                if manager.canvas.figure == figure
            ),
            None,
        )
        if num is not None:
            Gcf.destroy(num)



"""
Models representing entities in a plot, including containers (Figure, Axes) and
artists (Line, Grid, Image).

We follow the pattern that parents know about their children but children do
not know about their parents: thus, Figures know about their Axes and Axes know
about their Artists.
"""
import collections
import uuid as uuid_module

from ..utils.event import EmitterGroup, Event
from ..utils.list import EventedList
from ..utils.dict_view import UpdateOnlyDict, DictView


class BaseSpec:
    "Just a class with a uuid attribute and some slots."
    __slots__ = ("_uuid", "events", "__weakref__")

    def __init__(self, uuid):
        if uuid is None:
            uuid = uuid_module.uuid4()
        self._uuid = uuid

    @property
    def uuid(self):
        return self._uuid


class FigureSpec(BaseSpec):
    """
    Describes a Figure

    Parameters
    ----------
    axes : Tuple[AxesSpec]
    title : String
        Figure title text
    uuid : UUID, optional
        Automatically assigned to provide a unique identifier for this Figure,
        used internally to track it.
    short_title: String, optional
        Shorter figure title text, used in (for example) tab labels. Views
        should fall back on ``title`` if this is None.
    """

    __slots__ = ("_axes", "_title", "_short_title")

    def __init__(self, axes, *, title, uuid=None, short_title=None):
        for ax in axes:
            ax.set_figure(self)
        self._axes = tuple(axes)
        self._title = title
        self._short_title = short_title
        self.events = EmitterGroup(source=self, title=Event, short_title=Event)
        super().__init__(uuid)

    @property
    def axes(self):
        """
        Tuple of AxesSpecs. Set at FigureSpec creation time and immutable.

        Why is it immutable? Because rearranging Axes to make room for a new
        one is currently painful to do in matplotlib. This constraint might be
        relaxed in the future if the situation improves in matplotlib or if
        support for other plotting frameworks is added to bluesky-widgets.
        """
        return self._axes

    @property
    def title(self):
        "String for figure title. Settable."
        return self._title

    @title.setter
    def title(self, value):
        self._title = value
        self.events.title(value=value, figure_spec=self)

    @property
    def short_title(self):
        "String for figure title tab label. Settable"
        return self._short_title

    @short_title.setter
    def short_title(self, value):
        self._short_title = value
        self.events.short_title(value=value, figure_spec=self)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(axes={self.axes!r}, "
            f"title={self.title!r}, short_title={self.short_title!r}, "
            f"uuid={self.uuid!r})"
        )


class AxesSpec(BaseSpec):
    """
    Describes a set of Axes

    Paraemeters
    -----------
    lines : List[LineSpec], optional
    images : List[ImageSpec], optional
    title: String, optional
        Axes title text
    x_label : String, optional
        Text label for x axis
    y_label : String, optional
        Text label for y axis
    aspect : String or Float
        Passed through to :meth:`matplotlib.axes.Axes.imshow`
    x_limits : Tuple[Float]
        Limits of x axis
    y_limits : Tuple[Float]
        Limits of y axis
    uuid : UUID, optional
        Automatically assigned to provide a unique identifier for this Figure,
        used internally to track it.

    Examples
    --------

    Note that plot entities like lines may be declared at init time:

    >>> axes = AxesSpec(lines=[LineSpec(...)])

    Or added later:

    >>> axes = AxesSpec()
    >>> axes.lines.append(LineSpec(...))

    Or a mix:

    >>> axes = AxesSpec(lines=[LineSpec(...)])
    >>> axes.lines.append(LineSpec(...))

    And they can be removed at any point:

    >>> del axes.lines[0]  # Remove the first one.
    >>> del axes.lines[-1]  # Remove the last one.
    >>> axes.lines.clear()  # Remove them all.

    They may be accessed by type

    >>> axes.lines  # all lines
    [LineSpec(...), LineSpec(...), ...]

    Or by label

    >>> axes.by_label["Scan 8"]  # list of all plot entities with this label
    [LineSpec(...)]  # typically contains just one element
    """

    __slots__ = (
        "_type_map",
        "_figure",
        "_artists",
        "_lines",
        "_images",
        "_title",
        "_x_label",
        "_y_label",
        "_aspect",
        "_x_limits",
        "_y_limits",
    )

    def __init__(
        self,
        *,
        lines=None,
        images=None,
        title=None,
        x_label=None,
        y_label=None,
        aspect=None,
        x_limits=None,
        y_limits=None,
        uuid=None,
    ):
        self._figure = None
        self._lines = LineSpecList(lines or [])
        self._images = ImageSpecList(images or [])
        # A colleciton of all artists, mappping UUID to object
        self._artists = {}
        self._title = title
        self._x_label = x_label
        self._y_label = y_label
        self._aspect = aspect
        self._x_limits = x_limits
        self._y_limits = y_limits
        self.events = EmitterGroup(
            source=self,
            figure=Event,
            title=Event,
            x_label=Event,
            y_label=Event,
            aspect=Event,
            x_limits=Event,
            y_limits=Event,
        )
        super().__init__(uuid)
        for line in self._lines:
            self._adopt_artist(line)
        for image in self._images:
            self._adopt_artist(image)
        self._lines.events.added.connect(self._on_artist_added)
        self._lines.events.removed.connect(self._on_artist_removed)
        self._images.events.added.connect(self._on_artist_added)
        self._images.events.removed.connect(self._on_artist_removed)
        self._type_map = {
            LineSpec: self._lines,
            ImageSpec: self._images,
        }

    @property
    def figure(self):
        """
        The Figure in which this Axes is located.

        See Also
        --------
        :meth:`set_figure`
        """
        return self._figure

    def set_figure(self, figure):
        """
        This is called by FigureSpec when the Axes is added to it.

        It may only be called once.
        """
        if self._figure is not None:
            raise RuntimeError(
                f"Figure may only be set once. The Axes {self} already belongs "
                f"to {self.figure} and thus cannot be added to a new Figure."
            )
        self._figure = figure
        self.events.figure(value=figure)

    @property
    def lines(self):
        "List of LineSpecs on these Axes. Mutable."
        return self._lines

    @property
    def images(self):
        "List of ImageSpecs on these Axes. Mutable"
        return self._images

    @property
    def by_label(self):
        """
        Access artists as a read-only dict keyed by label.

        Since two artists are allowed to have the same label, the values are
        *lists*. In the common case, the list will have just one element.

        Examples
        --------

        Look up an object (e.g. a line) by its label and change its color.

        >>> spec = axes_spec.by_label["Scan 3"]
        >>> spec.style.update(color="red")
        """
        mapping = collections.defaultdict(list)
        for artist in self._artists.values():
            mapping[artist.label].append(artist)
        return DictView(dict(mapping))

    @property
    def by_uuid(self):
        """
        Access artists as a read-only dict keyed by uuid.
        """
        # Return a copy to prohibit mutation of internal bookkeeping.
        return DictView(self._artists)

    def discard(self, artist):
        "Discard any Aritst."
        try:
            self._type_map[type(artist)].remove(artist)
        except ValueError:
            pass

    def remove(self, artist):
        "Remove any Aritst."
        self._type_map[type(artist)].remove(artist)

    @property
    def title(self):
        "String for figure title. Settable"
        return self._title

    @title.setter
    def title(self, value):
        self._title = value
        self.events.title(value=value)

    @property
    def x_label(self):
        "String for x axes label. Settable."
        return self._x_label

    @x_label.setter
    def x_label(self, value):
        self._x_label = value
        self.events.x_label(value=value)

    @property
    def y_label(self):
        "String for y axes label. Settable."
        return self._y_label

    @y_label.setter
    def y_label(self, value):
        self._y_label = value
        self.events.y_label(value=value)

    @property
    def aspect(self):
        "String or Float for aspect. Settable"
        return self._aspect

    @aspect.setter
    def aspect(self, value):
        self._aspect = value
        self.events.aspect(value=value)

    @property
    def x_limits(self):
        "Float for limits of x axis. Settable"
        return self._x_limits

    @x_limits.setter
    def x_limits(self, value):
        self._x_limits = value
        self.events.x_limits(value=value)

    @property
    def y_limits(self):
        "Float for limits of y axis. Settable"
        return self._y_limits

    @y_limits.setter
    def y_limits(self, value):
        self._y_limits = value
        self.events.y_limits(value=value)

    def _on_artist_added(self, event):
        artist = event.item
        self._adopt_artist(artist)

    def _adopt_artist(self, artist):
        artist.set_axes(self)
        self._artists[artist.uuid] = artist

    def _on_artist_removed(self, event):
        artist = event.item
        del self._artists[artist.uuid]

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(lines={self.lines!r}, "
            f"images={self.images!r}, title={self.title!r},"
            f"x_label={self.x_label!r}, y_label={self.y_label!r}, "
            f"aspect={self.aspect!r}, x_limits={self.x_limits!r}, "
            f"y_limits={self.y_limits!r}, uuid={self.uuid!r})"
        )


class ArtistSpec(BaseSpec):
    """
    Describes the data, computation, and style for an artist (plot element)

    func : callable
        Expected signature::

            func(run: BlueskyRun)

        The expected return type varies by artist.
    run : BlueskyRun
        Contains data to be visualized.
    label : String
        Label used in legend and for lookup by label on AxesSpec.
    style : Dict, optional
        Options passed through to plotting framework, such as ``color`` or
        ``label``.
    axes : AxesSpec, optional
        This may be specified here or set later using :meth:`set_axes`. Once
        specified, it cannot be changed.
    uuid : UUID, optional
        Automatically assigned to provide a unique identifier for this Figure,
        used internally to track it.
    """

    __slots__ = ("_func", "_run", "_label", "_style", "_axes")

    def __init__(self, func, run, label, style=None, axes=None, uuid=None):
        self._func = func
        self._run = run
        self._label = label
        self._style = UpdateOnlyDict(style or {})
        self._axes = axes
        self.events = EmitterGroup(source=self, label=Event, style_updated=Event)
        # Re-emit updates. It's important to re-emit (not just pass through)
        # because the consumer will need access to self.
        self._style.events.updated.connect(
            lambda event: self.events.style_updated(
                update=event.update, artist_spec=self
            )
        )
        super().__init__(uuid)

    def set_axes(self, axes):
        """
        This is called by AxesSpec when the Artist is added to it.

        It may only be called once.
        """
        if self._axes is not None:
            raise RuntimeError(
                f"Axes may only be set once. The artist {self} already belongs "
                f"to {self.axes} and thus cannot be added to {axes}."
            )
        self._axes = axes

    @property
    def axes(self):
        """
        The Axes on which this Artist is drawn.

        See Also
        --------
        :meth:`set_axes`
        """
        return self._axes

    @property
    def func(self):
        "Function that transforms BlueskyRun to plottble data. Immutable."
        return self._func

    @property
    def run(self):
        "BlueskyRun that is the data source. Immutable."
        return self._run

    @property
    def label(self):
        "Label used in legend and for lookup by label on AxesSpec. Settable."
        return self._label

    @label.setter
    def label(self, value):
        self._label = str(value)
        self.events.label(value=value, artist_spec=self)

    @property
    def style(self):
        """
        Options passed to the artist.

        This *is* settable but it has to be done like:

        >>> spec.style.update({"color": "blue"})

        Attempts to modify the contents will be disallowed:

        >>> spec.style["color"] = blue  # TypeError!
        >>> del spec.style["color"]  # TypeError!
        """
        return self._style

    @style.setter
    def style(self, update):
        # Provide a more helpful error than the default,
        # AttributeError: can't set attribute.
        raise AttributeError(
            f"can't set attribute. Use style.update({update!r}) "
            f"instead of style = {update!r}."
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(func={self.func!r}, run={self.run!r}, "
            f"label={self.label!r}, style={self.style!r}, axes={self.axes!r}"
            f"uuid={self.uuid!r})"
        )


class LineSpec(ArtistSpec):
    """
    Describes the data, computation, and style for a line

    func : callable
        Expected signature::

            func(run: BlueskyRun) -> x: Array, y: Array

    run : BlueskyRun
        Contains data to be visualized.
    label : String
        Label used in legend and for lookup by label on AxesSpec.
    style : Dict, optional
        Options passed through to plotting framework, such as ``color`` or
        ``label``.
    axes : AxesSpec, optional
        This may be specified here or set later using :meth:`set_axes`. Once
        specified, it cannot be changed.
    uuid : UUID, optional
        Automatically assigned to provide a unique identifier for this Figure,
        used internally to track it.
    """

    __slots__ = ()


class ImageSpec(ArtistSpec):
    "Describes an image (both data and style)"


# EventedLists for each type of spec. We plan to add type-checking to these,
# hence a specific container for each.


class FigureSpecList(EventedList):
    __slots__ = ()


class AxesSpecList(EventedList):
    __slots__ = ()


class LineSpecList(EventedList):
    __slots__ = ()


class ImageSpecList(EventedList):
    __slots__ = ()



from bluesky_live.run_builder import build_simple_run

from .. import AutoLines
from ....headless.figures import HeadlessFigures


# Make some runs to use.
runs = [
    build_simple_run(
        {"motor": [1, 2], "det": [10, 20], "det2": [15, 25]},
        metadata={"scan_id": 1 + i},
    )
    for i in range(10)
]
MAX_RUNS = 3


def test_pinned():
    "Test AutoLines with 'pinned' and un-pinned runs."
    NUM_YS = 2
    model = AutoLines(max_runs=MAX_RUNS)
    view = HeadlessFigures(model.figures)
    assert not model.figures

    # Add MAX_RUNS and then some more and check that they do get bumped off.
    for run in runs[:5]:
        model.add_run(run)
        assert len(model.plot_builders[0].runs) <= MAX_RUNS
    assert runs[2:5] == list(model.plot_builders[0].runs)
    assert len(model.figures) == 1

    # Add a pinned run.
    pinned_run = runs[5]
    model.add_run(pinned_run, pinned=True)
    assert (
        frozenset([pinned_run.metadata["start"]["uid"]])
        == model.plot_builders[0].pinned
    )
    for run in runs[6:]:
        model.add_run(run)
        assert len(model.plot_builders[0].runs) == 1 + MAX_RUNS
        for axes_index in range(NUM_YS):
            assert len(model.figures[0].axes[axes_index].lines) == (1 + MAX_RUNS)
    # Check that it hasn't been bumped off.
    assert pinned_run in model.plot_builders[0].runs
    assert len(model.figures) == 1

    # Remove the pinned run.
    model.discard_run(pinned_run)
    assert len(model.plot_builders[0].runs) == MAX_RUNS
    for axes_index in range(NUM_YS):
        assert len(model.figures[0].axes[axes_index].lines) == MAX_RUNS
    assert pinned_run not in model.plot_builders[0].runs

    view.close()


def test_decrease_max_runs():
    "Decreasing max_runs should remove the runs and their associated lines."
    INITIAL_MAX_RUNS = 5
    model = AutoLines(max_runs=INITIAL_MAX_RUNS)
    view = HeadlessFigures(model.figures)
    for run in runs[:5]:
        model.add_run(run)
    assert len(model.plot_builders[0].runs) == INITIAL_MAX_RUNS
    assert len(model.figures[0].axes[0].lines) == INITIAL_MAX_RUNS
    # Decrease max_runs.
    model.max_runs = MAX_RUNS
    assert len(model.plot_builders[0].runs) == MAX_RUNS
    assert len(model.figures[0].axes[0].lines) == MAX_RUNS

    view.close()



from bluesky_live.run_builder import build_simple_run
import numpy

from ..plot_builders import Images
from ..plot_specs import AxesSpec, FigureSpec
from ...headless.figures import HeadlessFigure


def test_image():
    "Test Images with a 2D array."
    run = build_simple_run({"ccd": numpy.random.random((11, 13))})
    model = Images("ccd")
    view = HeadlessFigure(model.figure)
    assert not model.figure.axes[0].images
    model.add_run(run)
    assert model.figure.axes[0].images
    view.close()


def test_image_reduction():
    "Test Images with higher-dimensional arrays."
    dims = (5, 7, 11, 13, 17, 19)
    for i in range(3, len(dims)):
        run = build_simple_run({"ccd": numpy.random.random(dims[:i])})
    model = Images("ccd")
    view = HeadlessFigure(model.figure)
    model.add_run(run)
    view.close()


def test_properties():
    "Touch various accessors"
    run = build_simple_run({"ccd": numpy.random.random((11, 13))})
    model = Images("c * ccd", namespace={"c": 3})
    view = HeadlessFigure(model.figure)
    model.add_run(run)
    assert model.runs[0] is run
    assert model.field == "c * ccd"
    assert dict(model.namespace) == {"c": 3}
    assert model.needs_streams == ("primary",)
    view.close()


def test_figure_set_after_instantiation():
    axes = AxesSpec()
    model = Images("ccd", axes=axes)
    assert model.figure is None
    figure = FigureSpec((axes,), title="")
    assert model.figure is figure
    view = HeadlessFigure(model.figure)
    view.close()



import abc

from ..plot_specs import FigureSpecList
from ..utils import run_is_live_and_not_completed


class AutoPlotter(abc.ABC):
    """
    Generate best-effort figures based on a heuristic.

    Subclasses must define the method ``handle_new_stream``.

    Parameters
    ----------
    plot_suggestors : List[PlotSuggestors]
    max_runs : Integer, optional
        Default, ``None``, defers to the default of the ``plot_builder``.
    """

    def __init__(self):
        self.figures = FigureSpecList()
        self.figures.events.removed.connect(self._on_figure_removed)
        self.plot_builders = []

    def add_run(self, run, **kwargs):
        """
        Add a Run.

        Parameters
        ----------
        run : BlueskyRun
        **kwargs
            Passed through to plot_builder
        """
        for stream_name in run:
            self.handle_new_stream(run, stream_name, **kwargs)
        if run_is_live_and_not_completed(run):
            # Listen for additional streams.

            def pass_to_handle_new_stream(event):
                self.handle_new_stream(run, event.name, **kwargs)

            run.events.new_stream.connect(pass_to_handle_new_stream)
            # When run is complete, stop listening.
            run.events.completed.connect(
                lambda event: run.events.new_stream.disconnect(
                    pass_to_handle_new_stream
                )
            )

    def discard_run(self, run):
        """
        Discard a Run.

        If the Run is not present, this will return silently. Also,
        note that this only affect "active" plots that are currently
        receive new runs. Inactive ones will be left as they are.

        Parameters
        ----------
        run : BlueskyRun
        """
        for plot_builder in self.plot_builders:
            plot_builder.discard_run(run)

    @abc.abstractmethod
    def handle_new_stream(self, run, stream_name, **kwargs):
        "Build a plot, or add to an existing plot, or do nothing."

    def handle_figure_removed(self, figure):
        for plot_builder in list(self.plot_builders):
            if hasattr(plot_builder, "figure"):
                if figure is plot_builder.figure:
                    self.plot_builders.remove(plot_builder)
            elif hasattr(plot_builder, "figures"):
                for figure in self.figures:
                    if figure is plot_builder.figure:
                        self.plot_builders.remove(plot_builder)
            else:
                ValueError(
                    "A plot builder is expected to have an attribute "
                    "`figure` or `figures`"
                )

    def _on_figure_removed(self, event):
        self.handle_figure_removed(event.item)



import logging

from qtpy.QtCore import QStringListModel
from qtpy.QtWidgets import (
    QComboBox,
    QPushButton,
    QTabWidget,
    QVBoxLayout,
    QWidget,
    QSpacerItem,
    QSizePolicy,
)
from ._search_input import QtSearchInput
from ._search_results import QtSearchResults

logger = logging.getLogger(__name__)


class QtSubcatalogSelector(QComboBox):
    """
    ComboBox for selecting a subcatalog from a catalog-of-catalogs.
    """

    def __init__(self, names, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setModel(QStringListModel(names))
        self.setCurrentIndex(-1)


class QtSearch(QWidget):
    """
    A Qt view for a Search model.
    """

    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model

        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        # A "Back" button
        self._back_button = QPushButton("Back")
        self._back_button.setEnabled(False)
        self.layout().addWidget(self._back_button)
        self._back_button.clicked.connect(model.go_back)

        # Hook up model Events to Qt Slots.
        self.model.events.enter.connect(self.on_enter)
        self.model.events.go_back.connect(self.on_go_back)
        self.model.events.run_search_ready.connect(self.on_run_search_ready)
        self.model.events.run_search_cleared.connect(self.on_run_search_cleared)

        self._selector_widgets = []  # QComboBoxes
        self._run_search_widgets = []  # The SearchInput and SearchOutput widgets

        run_search = model.run_search
        if run_search:
            # The root catalog contains Runs, so immediately display Run Search
            # input and output.
            self._initialize_run_search(
                run_search.search_input, run_search.search_results
            )
            # No need to have a "Back" button in this case
            self._back_button.setVisible(False)
        else:
            # The root catalog is a catalog-of-catalogs, so display a combo
            # box.
            self._initialize_selector(list(model.current_catalog))

    def on_enter(self, event=None):
        "We are entering a subcatalog."
        names = list(event.catalog)
        self._initialize_selector(names)
        self._back_button.setEnabled(True)

    def _initialize_selector(self, names):
        "Create a combobox to select from subcatalogs."
        selector = QtSubcatalogSelector(names)
        self._selector_widgets.append(selector)

        def on_selection(index):
            name = names[index]
            try:
                self.model.enter(name)
            except Exception:
                logger.exception("Failed to select %r", name)
                selector.setCurrentIndex(-1)
            else:
                selector.setEnabled(False)

        selector.activated.connect(on_selection)
        self.layout().addWidget(selector)
        self.layout().addItem(
            QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Minimum)
        )

    def on_go_back(self, event):
        "Move up the tree of subcatalogs by one step."
        breadcrumbs = self.model.breadcrumbs
        while len(self._selector_widgets) > len(breadcrumbs) + 1:
            widget = self._selector_widgets.pop()
            widget.close()
        self._selector_widgets[-1].setEnabled(True)
        if not breadcrumbs:
            # This is the last widget. Disable back button.
            self._back_button.setEnabled(False)

    def on_run_search_ready(self, event):
        "We have a catalog of Runs."
        self._initialize_run_search(event.search_input, event.search_results)
        self._back_button.setEnabled(True)

    def _initialize_run_search(self, search_input, search_results):
        "Create search input and output for a catalog of Runs."
        # Create run search widgets and stash them as state for removal later.
        self._run_search_widgets.extend(
            [QtSearchInput(search_input), QtSearchResults(search_results)]
        )
        for w in self._run_search_widgets:
            self.layout().addWidget(w)

    def on_run_search_cleared(self, event):
        "Clear search input and output."
        for w in self._run_search_widgets:
            w.setParent(None)
        self._run_search_widgets.clear()


class QtSearches(QTabWidget):
    """
    Each tab is a QtSearch.
    """

    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # There is no way to *add* tabs in the UI, so for now we will not
        # permit removing them in the UI either. This could be added in the
        # future. In the meantime, it is supported from the model / console.
        self.setTabsClosable(False)
        self.tabCloseRequested.connect(self.close_tab)
        self.currentChanged.connect(self.on_current_changed)

        self.model = model
        self._tabs = {}  # map internal model to tab
        self.model.events.added.connect(self.on_added)
        self.model.events.removed.connect(self.on_removed)
        self.model.events.active.connect(self.on_active_changed)

    # These methods update the view in response to the SearchList model.

    def on_added(self, event):
        tab = QtSearch(event.item)
        self._tabs[event.item] = tab
        self.insertTab(event.index, tab, f"Search {event.item.name}")

    def on_removed(self, event):
        # A Search has been removed from the SearchList model.
        # Close the associated tab and clean up the associated state.
        widget = self._tabs[event.item]
        index = self.indexOf(widget)
        self.removeTab(index)
        del self._tabs[event.item]

    def on_active_changed(self, event):
        self.setCurrentWidget(self._tabs[event.item])

    # These methods notify the SearchList model to actions in the view.

    def on_current_changed(self, index):
        widget = self.widget(index)
        for model, tab in self._tabs.items():
            if tab is widget:
                widget.model.active = True
                break

    def close_tab(self, index):
        # Remove the associated Search model from the SearchList model.
        # Its removal will trigger on_removed above and thereby update the view.
        widget = self.widget(index)
        self.model.remove(widget.model)



"""
This module contains utility functions used by the built-in auto-plotters.
The module is private because the interfaces may change in the future.
Use the auto-plotters in bluesky_widgets.models.auto_plot_builders to use this
functionality.
"""


def hinted_fields(descriptor):
    "Which columns are the most important ones to show and visualize?"
    columns = []
    if descriptor["object_keys"]:
        # We will see if these objects hint at whether
        # a subset of their data keys ('fields') are interesting. If they
        # did, we'll use those. If these didn't, we know that the RunEngine
        # *always* records their complete list of fields, so we can use
        # them all unselectively.
        for obj_name, all_fields in descriptor["object_keys"].items():
            try:
                fields = descriptor.get("hints", {}).get(obj_name, {})["fields"]
            except KeyError:
                fields = all_fields
            columns.extend(fields)
    else:
        # There are no object_keys. This came from something other than the
        # RunEngine. Just use all the columns.
        columns.extend(descriptor["data_keys"])
    return columns



from bluesky_widgets.models.search import Search

headings = (
    "Unique ID",
    "Transient Scan ID",
    "Plan Name",
    "Start Time",
    "Duration",
    "Exit Status",
)


def extract_results_row_from_run(run):
    """
    Given a BlueskyRun, format a row for the table of search results.
    """
    from datetime import datetime

    metadata = run.describe()["metadata"]
    start = metadata["start"]
    stop = metadata["stop"]
    start_time = datetime.fromtimestamp(start["time"])
    if stop is None:
        str_duration = "-"
    else:
        duration = datetime.fromtimestamp(stop["time"]) - start_time
        str_duration = str(duration)
        str_duration = str_duration[: str_duration.index(".")]
    return (
        start["uid"][:8],
        start.get("scan_id", "-"),
        start.get("plan_name", "-"),
        start_time.strftime("%Y-%m-%d %H:%M:%S"),
        str_duration,
        "-" if stop is None else stop["exit_status"],
    )


columns = (headings, extract_results_row_from_run)


class AddSearchMixin:
    """Provide an add_search method that makes a Search.

    This purpose is to provide a top-level method that has a default column layout.
    """

    def add_search(self, catalog, columns=extract_results_row_from_run):
        """
        Add a new Search form.
        """
        search = Search(catalog, columns=(headings, extract_results_row_from_run))
        self.searches.append(search)

    @property
    def active_search(self):
        """
        Convenience for accessing the currently-active Search form.
        """
        return self.searches.active



from bluesky_live.bluesky_run import BlueskyRun, DocumentCache
import event_model


def stream_documents_into_runs(add_run):
    """
    Convert a flat stream of documents to "live" BlueskyRuns.

    Parameters
    ----------
    add_run : callable
        This will be called as ``add_run(run: BlueskyRun)`` each time a 'start'
        document is received.

    Returns
    -------
    callback : callable
        This should be subscribed to a callback registry that calls it like
        ``callback(name, doc)``.

    Examples
    --------

    This is used for connecting something that emits a flat stream of documents
    to something that wants to receive BlueskyRuns.

    Append to a plain list.

    >>> from bluesky import RunEngine
    >>> RE = RunEngine()
    >>> runs = []
    >>> RE.subscribe(stream_documents_into_runs(runs.append))

    Or, more usefully to an observable list.

    >>> from bluesky_widgets.models.utils import RunList
    >>> runs = RunList()
    >>> RE.subscribe(stream_documents_into_runs(runs.append))

    Add runs to a model with an ``add_run`` method. For example, it might be a
    model that generates figures.

    >>> from bluesky_widgets.models.plot_builders import AutoLines
    >>> model = AutoLines()

    >>> RE.subscribe(stream_documents_into_runs(model.add_run))
    """

    def factory(name, doc):
        dc = DocumentCache()

        def build_and_add_run(event):
            run = BlueskyRun(dc)
            add_run(run)

        dc.events.started.connect(build_and_add_run)
        return [dc], []

    rr = event_model.RunRouter([factory])
    return rr



def pytest_addoption(parser):
    """An option to show windows during tests. (Hidden by default).

    Showing windows decreases test speed by about %18.  Note, due to the
    placement of this conftest.py file, you must specify the bluesky_widgets
    folder (in the pytest command) to use this flag.

    Example
    -------
    $ pytest bluesky_widgets --show-window
    """
    parser.addoption(
        "--show-window",
        action="store_true",
        default=False,
        help="Show window during tests (not shown by default).",
    )



import tempfile
from suitcase.msgpack import Serializer
from bluesky import RunEngine
from bluesky.plans import count, scan, grid_scan
from ophyd.sim import det, motor, SynSignal, det4, motor1, motor2
import numpy as np

from databroker._drivers.msgpack import BlueskyMsgpackCatalog


random_img = SynSignal(func=lambda: np.random.random((5, 10, 10)), name="random_img")


def get_catalog():
    RE = RunEngine()

    directory = tempfile.TemporaryDirectory().name
    for i in range(1, 5):
        with Serializer(directory) as serializer:
            RE(scan([det], motor, -1, 1, 5 * i), serializer)
    with Serializer(directory) as serializer:
        RE(count([random_img], 3), serializer)
    with Serializer(directory) as serializer:
        RE(grid_scan([det4], motor1, -1, 2, 5, motor2, -1, 2, 7), serializer)

    catalog = BlueskyMsgpackCatalog(f"{directory}/*.msgpack")
    return catalog



"""
Use within IPython like

ipython --gui=qt

In [1]: %run -m bluesky_widgets.examples.ipy_qt_images
"""
from bluesky_widgets.examples.utils.generate_msgpack_data import get_catalog
from bluesky_widgets.models.auto_plot_builders import AutoImages
from bluesky_widgets.qt.figures import QtFigures

catalog = get_catalog()
counts = catalog.search({"plan_name": "count"})
run = counts[-1]


model = AutoImages()

model.add_run(run)
view = QtFigures(model.figures)
view.show()



"""
This file is vendored from pyFAI.app.calib2.

It has been modified to hook in bluesky_widgets' search functionality, available
from a new menu item in the top menu bar.

The changes are marked below with comments.
"""
#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Azimuthal integration
#             https://github.com/silx-kit/pyFAI
#
#    Copyright (C) 2017-2018 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""GUI tool for determining the geometry of a detector using a reference sample."""

__author__ = "Valentin Valls"
__contact__ = "valentin.valls@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "21/01/2020"
__status__ = "production"

import os
import sys
import datetime
from argparse import ArgumentParser

import logging

logging.basicConfig(level=logging.INFO)
logging.captureWarnings(True)
logger = logging.getLogger(__name__)
try:
    import hdf5plugin  # noqa
except ImportError:
    logger.debug("Unable to load hdf5plugin, backtrace:", exc_info=True)

logger_uncaught = logging.getLogger("pyFAI-calib2.UNCAUGHT")

import pyFAI.resources
import pyFAI.calibrant
import pyFAI.detectors
import pyFAI.io.image
from pyFAI.io.ponifile import PoniFile

try:
    from rfoo.utils import rconsole

    rconsole.spawn_server()
except ImportError:
    logger.debug("No socket opened for debugging. Please install rfoo")


def configure_parser_arguments(parser):
    # From AbstractCalibration

    parser.add_argument(
        "args", metavar="FILE", help="List of files to calibrate", nargs="*"
    )

    # Not yet used
    parser.add_argument(
        "-o",
        "--out",
        dest="outfile",
        help="Filename where processed image is saved",
        metavar="FILE",
        default=None,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="debug",
        action="store_true",
        default=False,
        help="switch to debug/verbose mode",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--opengl",
        "--gl",
        dest="opengl",
        action="store_true",
        default=False,
        help="Enable OpenGL rendering (else matplotlib is used)",
    )

    # Settings
    parser.add_argument(
        "-c",
        "--calibrant",
        dest="spacing",
        metavar="FILE",
        help="Calibrant name or file containing d-spacing of the reference sample (case sensitive)",
        default=None,
    )
    parser.add_argument(
        "-w",
        "--wavelength",
        dest="wavelength",
        type=float,
        help="wavelength of the X-Ray beam in Angstrom.",
        default=None,
    )
    parser.add_argument(
        "-e",
        "--energy",
        dest="energy",
        type=float,
        help="energy of the X-Ray beam in keV (hc=%skeV.A)." % pyFAI.units.hc,
        default=None,
    )
    parser.add_argument(
        "-P",
        "--polarization",
        dest="polarization_factor",
        type=float,
        default=None,
        help="polarization factor, from -1 (vertical) to +1 (horizontal),"
        + " default is None (no correction), synchrotrons are around 0.95",
    )
    parser.add_argument(
        "-D",
        "--detector",
        dest="detector_name",
        help="Detector name (instead of pixel size+spline)",
        default=None,
    )
    parser.add_argument(
        "-m",
        "--mask",
        dest="mask",
        help="file containing the mask (for image reconstruction)",
        default=None,
    )
    parser.add_argument(
        "-p", "--pixel", dest="pixel", help="size of the pixel in micron", default=None
    )
    parser.add_argument(
        "-s",
        "--spline",
        dest="spline",
        help="spline file describing the detector distortion",
        default=None,
    )

    parser.add_argument(
        "-n",
        "--pt",
        dest="npt",
        help="file with datapoints saved. Example: basename.npt",
        default=None,
    )

    # Not yet used
    parser.add_argument(
        "-i",
        "--poni",
        dest="poni",
        metavar="FILE",
        help="file containing the diffraction parameter (poni-file) [not used].",
        default=None,
    )
    # Not yet used
    parser.add_argument(
        "-b",
        "--background",
        dest="background",
        help="Automatic background subtraction if no value are provided [not used]",
        default=None,
    )
    # Not yet used
    parser.add_argument(
        "-d",
        "--dark",
        dest="dark",
        help="list of comma separated dark images to average and subtract [not used]",
        default=None,
    )
    # Not yet used
    parser.add_argument(
        "-f",
        "--flat",
        dest="flat",
        help="list of comma separated flat images to average and divide [not used]",
        default=None,
    )
    # Not yet used
    parser.add_argument(
        "--filter",
        dest="filter",
        help="select the filter, either mean(default), max or median [not used]",
        default=None,
    )

    # Geometry
    parser.add_argument(
        "-l",
        "--distance",
        dest="dist_mm",
        type=float,
        help="sample-detector distance in millimeter. Default: 100mm",
        default=None,
    )
    parser.add_argument(
        "--dist",
        dest="dist",
        type=float,
        help="sample-detector distance in meter. Default: 0.1m",
        default=None,
    )
    parser.add_argument(
        "--poni1",
        dest="poni1",
        type=float,
        help="poni1 coordinate in meter. Default: center of detector",
        default=None,
    )
    parser.add_argument(
        "--poni2",
        dest="poni2",
        type=float,
        help="poni2 coordinate in meter. Default: center of detector",
        default=None,
    )
    parser.add_argument(
        "--rot1",
        dest="rot1",
        type=float,
        help="rot1 in radians. default: 0",
        default=None,
    )
    parser.add_argument(
        "--rot2",
        dest="rot2",
        type=float,
        help="rot2 in radians. default: 0",
        default=None,
    )
    parser.add_argument(
        "--rot3",
        dest="rot3",
        type=float,
        help="rot3 in radians. default: 0",
        default=None,
    )
    # Constraints
    parser.add_argument(
        "--fix-wavelength",
        dest="fix_wavelength",
        help="fix the wavelength parameter. Default: Activated",
        default=True,
        action="store_true",
    )
    parser.add_argument(
        "--free-wavelength",
        dest="fix_wavelength",
        help="free the wavelength parameter. Default: Deactivated ",
        default=True,
        action="store_false",
    )
    parser.add_argument(
        "--fix-dist",
        dest="fix_dist",
        help="fix the distance parameter",
        default=None,
        action="store_true",
    )
    parser.add_argument(
        "--free-dist",
        dest="fix_dist",
        help="free the distance parameter. Default: Activated",
        default=None,
        action="store_false",
    )
    parser.add_argument(
        "--fix-poni1",
        dest="fix_poni1",
        help="fix the poni1 parameter",
        default=None,
        action="store_true",
    )
    parser.add_argument(
        "--free-poni1",
        dest="fix_poni1",
        help="free the poni1 parameter. Default: Activated",
        default=None,
        action="store_false",
    )
    parser.add_argument(
        "--fix-poni2",
        dest="fix_poni2",
        help="fix the poni2 parameter",
        default=None,
        action="store_true",
    )
    parser.add_argument(
        "--free-poni2",
        dest="fix_poni2",
        help="free the poni2 parameter. Default: Activated",
        default=None,
        action="store_false",
    )
    parser.add_argument(
        "--fix-rot1",
        dest="fix_rot1",
        help="fix the rot1 parameter",
        default=None,
        action="store_true",
    )
    parser.add_argument(
        "--free-rot1",
        dest="fix_rot1",
        help="free the rot1 parameter. Default: Activated",
        default=None,
        action="store_false",
    )
    parser.add_argument(
        "--fix-rot2",
        dest="fix_rot2",
        help="fix the rot2 parameter",
        default=None,
        action="store_true",
    )
    parser.add_argument(
        "--free-rot2",
        dest="fix_rot2",
        help="free the rot2 parameter. Default: Activated",
        default=None,
        action="store_false",
    )
    parser.add_argument(
        "--fix-rot3",
        dest="fix_rot3",
        help="fix the rot3 parameter",
        default=None,
        action="store_true",
    )
    parser.add_argument(
        "--free-rot3",
        dest="fix_rot3",
        help="free the rot3 parameter. Default: Activated",
        default=None,
        action="store_false",
    )

    parser.add_argument(
        "--npt",
        dest="npt_1d",
        help="Number of point in 1D integrated pattern, Default: 1024",
        type=int,
        default=None,
    )
    parser.add_argument(
        "--npt-azim",
        dest="npt_2d_azim",
        help="Number of azimuthal sectors in 2D integrated images. Default: 360",
        type=int,
        default=None,
    )
    parser.add_argument(
        "--npt-rad",
        dest="npt_2d_rad",
        help="Number of radial bins in 2D integrated images. Default: 400",
        type=int,
        default=None,
    )

    # Customization
    parser.add_argument(
        "--qtargs",
        dest="qtargs",
        type=str,
        help="Arguments propagated to Qt",
        default=None,
    )

    # Not yet used
    parser.add_argument(
        "--tilt",
        dest="tilt",
        help="Allow initially detector tilt to be refined (rot1, rot2, rot3). Default: Activated",
        default=None,
        action="store_true",
    )
    # Not yet used
    parser.add_argument(
        "--no-tilt",
        dest="tilt",
        help="Deactivated tilt refinement and set all rotation to 0",
        default=None,
        action="store_false",
    )
    # Not yet used
    parser.add_argument(
        "--saturation",
        dest="saturation",
        help="consider all pixel>max*(1-saturation) as saturated and "
        "reconstruct them, default: 0 (deactivated)",
        default=0,
        type=float,
    )
    # Not yet used
    parser.add_argument(
        "--weighted",
        dest="weighted",
        help="weight fit by intensity, by default not.",
        default=False,
        action="store_true",
    )
    # Not yet used
    parser.add_argument(
        "--unit",
        dest="unit",
        help="Valid units for radial range: 2th_deg, 2th_rad, q_nm^-1,"
        " q_A^-1, r_mm. Default: 2th_deg",
        type=str,
        default="2th_deg",
    )
    # Not yet used
    parser.add_argument(
        "--no-gui",
        dest="gui",
        help="force the program to run without a Graphical interface",
        default=True,
        action="store_false",
    )
    # Not yet used
    parser.add_argument(
        "--no-interactive",
        dest="interactive",
        help="force the program to run and exit without prompting" " for refinements",
        default=True,
        action="store_false",
    )

    # From Calibration
    # Not yet used
    parser.add_argument(
        "-r",
        "--reconstruct",
        dest="reconstruct",
        help="Reconstruct image where data are masked or <0  (for Pilatus "
        + "detectors or detectors with modules)",
        action="store_true",
        default=False,
    )
    # Not yet used
    parser.add_argument(
        "-g",
        "--gaussian",
        dest="gaussian",
        help="""Size of the gaussian kernel.
Size of the gap (in pixels) between two consecutive rings, by default 100
Increase the value if the arc is not complete;
decrease the value if arcs are mixed together.""",
        default=None,
    )
    # Not yet used
    parser.add_argument(
        "--square",
        dest="square",
        action="store_true",
        help="Use square kernel shape for neighbor search instead of diamond shape",
        default=False,
    )


description = """Calibrate the diffraction setup geometry based on
Debye-Sherrer rings images without a priori knowledge of your setup.
You will need to provide a calibrant or a "d-spacing" file containing the
spacing of Miller plans in Angstrom (in decreasing order).
%s or search in the American Mineralogist database:
http://rruff.geo.arizona.edu/AMS/amcsd.php""" % str(
    pyFAI.calibrant.ALL_CALIBRANTS
)

epilog = """The output of this program is a "PONI" file containing the
detector description and the 6 refined parameters (distance, center, rotation)
and wavelength. An 1D and 2D diffraction patterns are also produced.
(.dat and .azim files)"""


def parse_pixel_size(pixel_size):
    """Convert a comma separated sting into pixel size

    :param str pixel_size: String containing pixel size in micron
    :rtype: Tuple[float,float]
    :returns: Returns floating point pixel size in meter
    """
    sp = pixel_size.split(",")
    if len(sp) >= 2:
        try:
            result = [float(i) * 1e-6 for i in sp[:2]]
        except Exception:
            logger.error("Error in reading pixel size_2")
            raise ValueError("Not a valid pixel size")
    elif len(sp) == 1:
        px = sp[0]
        try:
            result = [float(px) * 1e-6, float(px) * 1e-6]
        except Exception:
            logger.error("Error in reading pixel size_1")
            raise ValueError("Not a valid pixel size")
    else:
        logger.error("Error in reading pixel size_0")
        raise ValueError("Not a valid pixel size")
    return result


def parse_options():
    """
    Returns parsed command line argument as an `options` object.

    :raises ExitException: In case of the use of `--help` in the comman line
    """
    usage = "pyFAI-calib2 [options] input_image.edf"
    parser = ArgumentParser(usage=usage, description=description, epilog=epilog)
    version = "calibration from pyFAI  version %s: %s" % (pyFAI.version, pyFAI.date)
    parser.add_argument("-V", "--version", action="version", version=version)
    configure_parser_arguments(parser)

    # Analyse aruments and options
    options = parser.parse_args()
    return options


def displayExceptionBox(message, exc_info):
    """
    Display an exception as a MessageBox

    :param str message: A context message
    :param Union[tuple,Exception] exc_info: An exception or the output of
        exc_info.
    """
    from pyFAI.gui.dialog import MessageBox

    MessageBox.exception(None, message, exc_info, logger)


def setup_model(model, options):
    """
    Setup the model using options from the command line.

    :param pyFAI.gui.model.CalibrationModel model: Model of the application
    """
    args = options.args

    # The module must not import the GUI
    from pyFAI.gui.utils import units

    # TODO: This should be removed
    import pyFAI.gui.cli_calibration

    # Settings
    settings = model.experimentSettingsModel()
    if options.spacing:
        calibrant = None
        try:
            if options.spacing in pyFAI.calibrant.CALIBRANT_FACTORY:
                calibrant = pyFAI.calibrant.CALIBRANT_FACTORY(options.spacing)
            elif os.path.isfile(options.spacing):
                calibrant = pyFAI.calibrant.Calibrant(options.spacing)
            else:
                logger.error("No such Calibrant / d-Spacing file: %s", options.spacing)
        except Exception as e:
            calibrant = None
            displayExceptionBox("Error while loading the calibrant", e)
        if calibrant:
            settings.calibrantModel().setCalibrant(calibrant)

    if options.wavelength:
        value = units.convert(
            options.wavelength, units.Unit.ANGSTROM, units.Unit.METER_WL
        )
        settings.wavelength().setValue(value)

    if options.energy:
        value = units.convert(options.energy, units.Unit.ENERGY, units.Unit.METER_WL)
        settings.wavelength().setValue(value)

    if options.polarization_factor:
        settings.polarizationFactor(options.polarization_factor)

    if options.detector_name:
        try:
            detector = pyFAI.gui.cli_calibration.get_detector(options.detector_name)
            if options.pixel:
                logger.warning(
                    "Detector model already specified. Pixel size argument ignored."
                )
        except Exception as e:
            detector = None
            displayExceptionBox("Error while loading the detector", e)
    elif options.pixel:
        pixel_size = parse_pixel_size(options.pixel)
        detector = pyFAI.detectors.Detector(pixel1=pixel_size[0], pixel2=pixel_size[0])
    else:
        detector = None

    if options.spline:
        try:
            if detector is None:
                detector = pyFAI.detectors.Detector(splineFile=options.spline)
            elif detector.__class__ is pyFAI.detectors.Detector or detector.HAVE_TAPER:
                detector.set_splineFile(options.spline)
            else:
                logger.warning(
                    "Spline file not supported with this kind of detector. Argument ignored."
                )
        except Exception as e:
            displayExceptionBox("Error while loading the spline file", e)

    geometryFromPoni = None
    if options.poni:
        settings.poniFile().setValue(options.poni)
        settings.poniFile().setSynchronized(True)

        if os.path.exists(options.poni):
            if detector is None:
                poniFile = PoniFile()
                poniFile.read_from_file(options.poni)
                detector = poniFile.detector

                from pyFAI.gui.model.GeometryModel import GeometryModel

                geometryFromPoni = GeometryModel()
                geometryFromPoni.distance().setValue(poniFile.dist)
                geometryFromPoni.poni1().setValue(poniFile.poni1)
                geometryFromPoni.poni2().setValue(poniFile.poni2)
                geometryFromPoni.rotation1().setValue(poniFile.rot1)
                geometryFromPoni.rotation2().setValue(poniFile.rot2)
                geometryFromPoni.rotation3().setValue(poniFile.rot3)
                geometryFromPoni.wavelength().setValue(poniFile.wavelength)

                geometryHistory = model.geometryHistoryModel()
                now = datetime.datetime.now()
                geometryHistory.appendGeometry("Ponifile", now, geometryFromPoni, None)
            else:
                logger.warning(
                    "Detector redefined in the command line. Detector from --poni argument ignored."
                )
        else:
            logger.warning(
                "PONI file '%s' do not exists. --poni option ignored.", options.poni
            )

    settings.detectorModel().setDetector(detector)

    if options.mask:
        try:
            with settings.mask().lockContext() as image_model:
                image_model.setFilename(options.mask)
                data = pyFAI.io.image.read_image_data(options.mask)
                image_model.setValue(data)
                image_model.setSynchronized(True)
        except Exception as e:
            displayExceptionBox("Error while loading the mask", e)

    if len(args) == 0:
        pass
    elif len(args) == 1:
        image_file = args[0]
        try:
            with settings.image().lockContext() as image_model:
                image_model.setFilename(image_file)
                data = pyFAI.io.image.read_image_data(image_file)
                image_model.setValue(data)
                image_model.setSynchronized(True)
        except Exception as e:
            displayExceptionBox("Error while loading the image", e)
    else:
        logger.error("Too much images provided. Only one is expected")

    # Geometry
    # FIXME it will not be used cause the fitted geometry will be overwrited
    geometry = model.fittedGeometry()
    if (
        options.dist_mm is not None
        or options.dist is not None
        or options.poni1 is not None
        or options.poni2 is not None
        or options.rot1 is not None
        or options.rot2 is not None
        or options.rot3 is not None
    ):
        if geometryFromPoni is not None:
            logger.warning(
                "Geometry redefined in the command line. Geometry from --poni argument was stored in the history."
            )
        if options.dist_mm is not None:
            geometry.distance().setValue(1e-3 * options.dist_mm)
        if options.dist is not None:
            geometry.distance().setValue(options.dist)
        if options.poni1 is not None:
            geometry.poni1().setValue(options.poni1)
        if options.poni2 is not None:
            geometry.poni2().setValue(options.poni2)
        if options.rot1 is not None:
            geometry.rotation1().setValue(options.rot1)
        if options.rot2 is not None:
            geometry.rotation2().setValue(options.rot2)
        if options.rot3 is not None:
            geometry.rotation3().setValue(options.rot3)
    else:
        if geometryFromPoni is not None:
            geometry.setFrom(geometryFromPoni)

    # Constraints
    constraints = model.geometryConstraintsModel()
    if options.fix_wavelength is not None:
        constraints.wavelength().setFixed(options.fix_wavelength)
    if options.fix_dist is not None:
        constraints.distance().setFixed(options.fix_dist)
    if options.fix_poni1 is not None:
        constraints.poni1().setFixed(options.fix_poni1)
    if options.fix_poni2 is not None:
        constraints.poni2().setFixed(options.fix_poni2)
    if options.fix_rot1 is not None:
        constraints.rotation1().setFixed(options.fix_rot1)
    if options.fix_rot2 is not None:
        constraints.rotation2().setFixed(options.fix_rot2)
    if options.fix_rot3 is not None:
        constraints.rotation3().setFixed(options.fix_rot3)

    integrationSettingsModel = model.integrationSettingsModel()
    npt = None
    if options.npt_1d is not None:
        npt = options.npt_1d
    if options.npt_2d_rad is not None:
        if npt is not None:
            logger.error("Both --npt and --npt-rad defined. The biggest is used.")
            npt = max(npt, options.npt_2d_rad)

    if npt is not None:
        integrationSettingsModel.nPointsRadial().setValue(npt)
    else:
        integrationSettingsModel.nPointsRadial().setValue(1024)

    if options.npt_2d_azim is not None:
        integrationSettingsModel.nPointsAzimuthal().setValue(options.npt_2d_azim)
    else:
        integrationSettingsModel.nPointsAzimuthal().setValue(360)

    # Integration
    if options.unit:
        unit = pyFAI.units.to_unit(options.unit)
        integrationSettingsModel.radialUnit().setValue(unit)

    if options.outfile:
        logger.error("outfile option not supported")

    if options.reconstruct:
        logger.error("reconstruct option not supported")
    if options.gaussian:
        logger.error("gaussian option not supported")
    if options.square:
        logger.error("square option not supported")
    if options.pixel:
        logger.error("pixel option not supported")

    if options.background:
        logger.error("background option not supported")
    if options.dark:
        logger.error("dark option not supported")
    if options.flat:
        logger.error("flat option not supported")
    if options.filter:
        logger.error("filter option not supported")

    if options.tilt:
        logger.error("tilt option not supported")
    if options.saturation:
        logger.error("saturation option not supported")
    if options.weighted:
        logger.error("weighted option not supported")

    if options.gui is not True:
        logger.error("gui option not supported")
    if options.interactive is not True:
        logger.error("interactive option not supported")

    if options.npt:
        try:
            from pyFAI.gui.helper import model_transform
            from pyFAI.control_points import ControlPoints

            controlPoints = ControlPoints(filename=options.npt)
            peakSelectionModel = model.peakSelectionModel()
            model_transform.initPeaksFromControlPoints(
                peakSelectionModel, controlPoints
            )
        except Exception as e:
            displayExceptionBox("Error while loading control-point file", e)


def logUncaughtExceptions(exceptionClass, exception, stack):
    try:
        import traceback

        if stack is not None:
            # Mimic the syntax of the default Python exception
            message = "".join(traceback.format_tb(stack))
            message = "{1}\nTraceback (most recent call last):\n{2}{0}: {1}".format(
                exceptionClass.__name__, exception, message
            )
        else:
            # There is no backtrace
            message = "{0}: {1}".format(exceptionClass.__name__, exception)
        logger_uncaught.error(message)
    except Exception:
        # Make sure there is no problem at all in this function
        try:
            logger_uncaught.error(exception)
        except Exception:
            print("Error:" + str(exception))


def main():
    # It have to be done before loading Qt
    # --help must also work without Qt
    options = parse_options()

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    # Then we can load Qt
    import silx
    from silx.gui import qt

    if options.opengl:
        silx.config.DEFAULT_PLOT_BACKEND = "opengl"

    # Make sure matplotlib is loaded first by silx
    import silx.gui.plot.matplotlib
    from pyFAI.gui.CalibrationWindow import CalibrationWindow
    from pyFAI.gui.CalibrationContext import CalibrationContext

    sys.excepthook = logUncaughtExceptions
    if options.qtargs is None:
        qtArgs = []
    else:
        qtArgs = options.qtargs.split()
    app = qt.QApplication(qtArgs)
    pyFAI.resources.silx_integration()

    settings = qt.QSettings(
        qt.QSettings.IniFormat, qt.QSettings.UserScope, "pyfai", "pyfai-calib2", None
    )

    context = CalibrationContext(settings)
    context.restoreSettings()

    setup_model(context.getCalibrationModel(), options)
    window = CalibrationWindow(context)

    # Begin modifications for bluesky_widgets

    from qtpy.QtWidgets import QDialog
    from bluesky_widgets.models.search.searches import Search
    from bluesky_widgets.qt.searches import QtSearch
    from bluesky_widgets.examples.utils.generate_msgpack_data import get_catalog
    from bluesky_widgets.examples.utils.add_search_mixin import columns

    from qtpy.QtWidgets import QAction, QHBoxLayout, QPushButton, QVBoxLayout, QWidget

    example_catalog = get_catalog()

    class SearchDialog(QDialog):
        """
        Combine the QtSearches widget with a button that processes selected Runs.
        """

        def __init__(self, model, *args, **kwargs):
            super().__init__(*args, **kwargs)
            layout = QVBoxLayout()
            self.setLayout(layout)
            self.setModal(True)
            layout.addWidget(QtSearch(model))

            accept = QPushButton("Open")
            reject = QPushButton("Cancel")
            # We'll just slip this into an existing widget --- not great form, but
            # this is just a silly example.
            accept.clicked.connect(self.accept)
            reject.clicked.connect(self.reject)
            buttons_widget = QWidget()
            buttons_layout = QHBoxLayout()
            buttons_widget.setLayout(buttons_layout)
            buttons_layout.addWidget(accept)
            buttons_layout.addWidget(reject)
            layout.addWidget(buttons_widget)

    def launch_dialog():
        model = Search(example_catalog, columns=columns)
        dialog = SearchDialog(model)

        def open_catalog():
            catalog = model.selection_as_catalog
            # TODO Extract the data from this.
            print(f"Open {len(catalog)} runs")

        dialog.accepted.connect(open_catalog)
        dialog.exec()

    menu = window.menuBar().addMenu("&Data Broker")
    launch_dialog_action = QAction("&Search")
    menu.addAction(launch_dialog_action)
    launch_dialog_action.triggered.connect(launch_dialog)
    # End modifications for bluesky_widgets

    window.setVisible(True)
    window.setAttribute(qt.Qt.WA_DeleteOnClose, True)

    result = app.exec_()
    context.saveSettings()

    # remove ending warnings relative to QTimer
    app.deleteLater()

    return result


if __name__ == "__main__":
    result = main()
    sys.exit(result)



from ..utils.event import Event, EmitterGroup


class RunTree:
    """
    Tree viewer for the metadata in a BlueskyRun.

    Parameters
    ----------
    run: BlueskyRun or None, optional

    """

    def __init__(self, run=None):
        # This is a very light model. It only has one piece of state and one
        # signal when that state changes.
        self._run = run
        self.events = EmitterGroup(source=self, run=Event)

    @property
    def run(self):
        "The currently-viewed Run. (None if empty.)"
        return self._run

    @run.setter
    def run(self, run):
        self._run = run
        self.events.run(run=run)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.run!r})"



import pytest
from ...examples.advanced.qt_viewer_with_search import ExampleApp


@pytest.fixture(scope="function")
def make_test_app(qtbot, request):
    apps = []

    def actual_factory(*model_args, **model_kwargs):
        model_kwargs["show"] = model_kwargs.pop(
            "show", request.config.getoption("--show-window")
        )
        app = ExampleApp(*model_args, **model_kwargs)
        apps.append(app)
        return app

    yield actual_factory

    for app in apps:
        app.close()


def test_app(make_test_app):
    "An integration test"
    app = make_test_app()
    assert not app.viewer.figures
    app.searches[0].input.query = {"plan_name": "scan"}
    # TODO It would be nice to hit the "View Selected Runs" button here, but
    # changing *selection* from the model is not supported.
    # https://github.com/bluesky/bluesky-widgets/issues/41
    catalog = app.searches[0].run_search.search_results.catalog
    assert len(catalog)  # Results should be non-empty
    for _, run in catalog.items():
        app.viewer.add_run(run)
    assert app.viewer.figures
    assert app.viewer.figures[0].axes[0].lines
    app.viewer.figures.clear()
